# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import List, Union

import dask
import dask.dataframe as dd
import numpy as np
import pandas as pd
from dask.base import normalize_token, tokenize
from openai import OpenAI

from nemo_curator.filters.doc_filter import DocumentFilter
from nemo_curator.utils.decorators import batched
from nemo_curator.utils.distributed_utils import NoWorkerError, load_object_on_worker


def create_client(base_url, api_key):
    openai_client = OpenAI(
        base_url=base_url,
        api_key=api_key,
    )
    return openai_client


# ----------------------------------------------------------------------------80
# ------------------------ EASINESS FILTER -------------------------------------
# ----------------------------------------------------------------------------80
class EasinessFilter(DocumentFilter):
    """
    Discards questions that are deemed easy to retrieve by retriever modls
    """

    def __init__(
        self,
        base_url: str,
        api_key: str,
        model: str,
        percentile: float = 0.7,
        truncate: str = "NONE",
        batch_size: int = 1,
        text_fields: List[str] = ["text", "question"],
    ):

        self._name = "easiness_filter"
        self.base_url = base_url
        self.api_key = api_key
        self.nim_model = model
        self.percentile = percentile
        if truncate:
            self.truncate = truncate
        self.batch_size = batch_size
        self.text_fields = text_fields

    @batched
    def score_document(self, df: pd.DataFrame):

        try:
            self.client = load_object_on_worker(
                attr="openai_client_easiness",
                load_object_function=create_client,
                load_object_kwargs={"base_url": self.base_url, "api_key": self.api_key},
            )
        except NoWorkerError:
            return pd.Series(np.ones(len(df)), dtype=float)

        document_score = self._calc_similarity_nim(
            df[self.text_fields[0]].to_list(), df[self.text_fields[1]].to_list()
        )
        return pd.Series(document_score, index=df.index)

    @batched
    def keep_document(self, scores: pd.Series):
        filter_threshold = np.percentile(scores, self.percentile)
        return scores <= filter_threshold

    def _get_nim_embedding(self, text, input_type):
        # Obtain embeddings from nim model
        if isinstance(text, list):
            input_ = text
        elif isinstance(text, str):
            input_ = [text]

        try:
            response = self.client.embeddings.create(
                input=input_,
                model=self.nim_model,
                encoding_format="float",
                extra_body={"input_type": input_type, "truncate": self.truncate},
            )
        except Exception as e:
            print(f"Error: {e}")
            response = None

        if response and not isinstance(response, str):
            if isinstance(text, list):
                embeddings = [r.embedding for r in response.data]
            elif isinstance(text, str):
                embeddings = response.data[0].embedding
            return embeddings
        else:
            return []

    def _calc_similarity_nim(self, context, question):
        # cosine similarity
        doc_embed = self._get_nim_embedding(text=context, input_type="passage")
        q_embed = self._get_nim_embedding(text=question, input_type="query")
        if isinstance(context, list) and isinstance(question, list):
            if doc_embed and q_embed:
                sim = np.diag(np.dot(np.array(doc_embed), np.array(q_embed).T))
            else:
                sim = np.zeros(len(context))
        else:
            if doc_embed and q_embed:
                sim = np.dot(doc_embed, q_embed)
            else:
                sim = 0.0

        return sim


# ----------------------------------------------------------------------------80
# ----------------------- Answerability Filter ---------------------------------
# ----------------------------------------------------------------------------80


class AnswerabilityFilter(DocumentFilter):
    """
    Discards questions that are not answerable by content present in the
    context document
    """

    def __init__(
        self,
        base_url: str,
        api_key: str,
        model: str,
        answerability_system_prompt: str,
        answerability_user_prompt_template: str,
        num_criteria: int,
        text_fields: List[str] = ["text", "question"],
    ):

        self._name = "answerability_filter"
        self.base_url = base_url
        self.api_key = api_key
        self.model_name = model
        self.system_prompt = answerability_system_prompt
        self.user_prompt_template = answerability_user_prompt_template
        self.num_criteria = num_criteria
        self.text_fields = text_fields

    @batched
    def score_document(self, df: pd.DataFrame):

        try:
            self.client = load_object_on_worker(
                attr="openai_client_answerability",
                load_object_function=create_client,
                load_object_kwargs={"base_url": self.base_url, "api_key": self.api_key},
            )
        except NoWorkerError:
            return pd.Series(["string"] * len(df))

        return df.progress_apply(
            lambda row: self._llm_as_judge(
                row[self.text_fields[0]],
                row[self.text_fields[1]],
            ),
            axis=1,
        )

    # ----------------------------------------------------------------------------80
    @batched
    def keep_document(self, scores: pd.Series):

        def _keep_document(score: str):
            is_keep = True  # default is to keep
            try:
                json_ans = json.loads(score)
                for i in range(self.num_criteria):
                    if json_ans[f"criterion_{i+1}"] != "Y":
                        # filter out data if any of the criteria fails
                        is_keep = False  # filter out
                        break
            except Exception as e:
                pass  # TODO log the errors
                # print(f"Parse error {e}")
                # if there is a parse error, keep the document

            return is_keep

        return scores.apply(_keep_document)

    def _llm_as_judge(self, context: str, question: str):

        user_query = self.system_prompt + "\n\n"
        user_query += self.user_prompt_template.format(
            context=context, question=question
        )

        try:
            completion = self.client.chat.completions.create(
                model=self.model_name,
                messages=[{"role": "user", "content": user_query}],
                temperature=0.5,
                top_p=1,
                max_tokens=1024,
            )

            generation = completion.choices[0].message.content

        except Exception as e:
            print(f"API call error {e}")
            return None  # generation

        return generation


# ----------------------------------------------------------------------------80



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import os
import string
import tokenize
import warnings
from io import StringIO
from itertools import groupby


def get_word_splitter(language):
    """
    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.

    Args:
        language (str): An ISO 639-1 language code.
            For example, "en" for English, "zh" for Chinese, and "ja" for Japanese.
    Returns:
        A function which can be used to parse the words of a string into a list.
    """
    language = language.lower()

    if language == "zh":
        # We use the Jieba library which is a Chinese word segmentation module
        # because Chinese text is not separated by spaces.
        import jieba

        def jieba_splitter(text):
            return list(jieba.cut(text))

        return jieba_splitter

    elif language == "ja":
        # We use the MeCab library which is a morphological analyzer for Japanese text
        # because Japanese text is not separated by spaces.
        import MeCab

        def mecab_splitter(text):
            mecab = MeCab.Tagger()
            parsed = mecab.parse(text)
            lines = parsed.strip().split("\n")
            tokens = [line.split("\t")[0] for line in lines if line and line != "EOS"]
            return tokens

        return mecab_splitter

    else:

        def default_splitter(text):
            return text.split()

        return default_splitter


def get_paragraphs(document):
    # Split the document into paragraphs.
    # A paragraph is defined as a sequence of lines
    # separated by a double newline.
    return document.split("\n\n")


def get_sentences(document):
    # Split the document into sentences.
    # A sentence is defined as a sequence of lines separated
    # by a single newline.
    return [x for x in document.split("\n") if len(x.strip()) > 0]


def get_ngrams(input_list, n):
    # Fast function to return n-grams from a list of tokens.
    return [item for item in zip(*[input_list[i:] for i in range(n)])]


def is_paragraph_indices_in_top_or_bottom_only(
    boilerplate_paragraph_indices,
    num_paragraphs,
):

    def _is_contiguous(indices):
        # Indices are sorted in ascending order.
        num_indices = len(indices) - 1
        return all(indices[i] + 1 == indices[i + 1] for i in range(num_indices))

    # See if the indices are contiguous and exclusively at the top/bottom.
    # Indices are sorted in ascending order.
    # If num_paragraphs = 11:
    # Valid indices example : [0, 1, 9, 10]
    # Invalid indices example : [0, 1, 3, 9, 10]
    # Invalid indices example : [0, 1, 3, 5, 6, 9, 10]
    # Invalid indices example : [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    if len(boilerplate_paragraph_indices) == num_paragraphs:
        return False
    return _is_contiguous(boilerplate_paragraph_indices) and (
        boilerplate_paragraph_indices[0] == 0
        or boilerplate_paragraph_indices[-1] == num_paragraphs - 1
    )


# Node types for processing abstract syntax tree
NODE_TYPES = {
    ast.ClassDef: "Class",
    ast.FunctionDef: "Function/Method",
    ast.Module: "Module",
}


def get_comments_and_docstring(source, comments=True, clean_comments=False):
    """
    Extract all natural text in source: comments + doctsrings
      the extraction fails in case of syntax errors in the file
      Args:
          source: the code to parse
          comments: if True extract comments two
          clean_comment: if True remove # from extracted comments
      Returns:
          a string with concatenated docstrings and comments
    """

    try:
        docstrings = "\n".join(get_docstrings(source))
    except Exception:
        docstrings = None
        warnings.warn(
            "code couldn't be parsed due to compilation failure, "
            "no docstring is extracted"
        )

    if comments:
        try:
            comments = get_comments(source, clean=clean_comments)
        except Exception:
            comments = None
            warnings.warn("tokenization error, no comment is extracted")
    else:
        comments = ""

    return docstrings, comments


def get_comments(s, clean=False):
    "Returns a string including all coments"
    coments = []
    g = tokenize.generate_tokens(StringIO(s).readline)
    for toknum, tokval, _, _, _ in g:
        # print(toknum,tokval)
        if toknum == tokenize.COMMENT:
            coments.append((toknum, tokval))
    result = tokenize.untokenize(coments)
    if clean:
        result = result.replace("#", "")
    return result


def get_docstrings(source, module="<string>"):
    """Parse Python source code from file or string and print docstrings."""
    if hasattr(source, "read"):
        filename = getattr(source, "name", module)
        module = os.path.splitext(os.path.basename(filename))[0]
        source = source.read()

    docstrings = sorted(
        parse_docstrings(source), key=lambda x: (NODE_TYPES.get(type(x[0])), x[1])
    )

    grouped = groupby(docstrings, key=lambda x: NODE_TYPES.get(type(x[0])))
    results = []
    for _, group in grouped:
        for _, name, docstring in group:
            name = name if name else module
            if docstring:
                results.append(docstring)
    return results


def parse_docstrings(source):
    """Parse Python source code and yield a tuple of ast node instance, name,
    and docstring for each function/method, class and module."""
    tree = ast.parse(source)

    for node in ast.walk(tree):
        if isinstance(node, tuple(NODE_TYPES)):
            docstring = ast.get_docstring(node)

            yield (node, getattr(node, "name", None), docstring)


def remove_punctuation(str_in):
    return str_in.translate(str_in.maketrans("", "", string.punctuation))


def get_words(text):
    word_start_char_positions = []
    prev = 0
    words = []

    text = text.lower()
    text = remove_punctuation(text)
    if len(text) > 0:
        for i in range(len(text)):
            if text[i] != " ":
                if i == 0 or text[i - 1] == " ":
                    word_start_char_positions.append(i)
                    if i != 0:
                        words.append(text[prev:i].strip())
                    prev = i
        words.append(text[prev : i + 1].strip())
        if words[0] == "":
            words = words[1:]
    return words, word_start_char_positions



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pydoc import locate

import yaml

import nemo_curator
from nemo_curator.download.doc_builder import (
    import_downloader,
    import_extractor,
    import_iterator,
)
from nemo_curator.filters import import_filter
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir


def build_filter(filter_config):
    # Import the filter
    filter_class = import_filter(filter_config["name"])

    # Check if constructor has been provided
    if ("params" not in filter_config) or (filter_config["params"] is None):
        filter_config["params"] = {}

    doc_filter = filter_class(**filter_config["params"])

    if filter_config.get("filter_only", False):
        filter_stage = nemo_curator.Filter(
            doc_filter.keep_document, filter_field=doc_filter.name
        )
    else:
        score_field = (
            doc_filter._name if filter_config.get("log_score", False) else None
        )
        filter_stage = nemo_curator.ScoreFilter(
            doc_filter, filter_config.get("input_field"), score_field=score_field
        )

    return filter_stage


def build_filter_pipeline(filter_config_file):
    # Get the filter config file
    with open(filter_config_file, "r") as config_file:
        filter_params = yaml.load(config_file, Loader=yaml.FullLoader)

    filters = []
    text_field = filter_params.get("input_field")
    for nc_filter_config in filter_params.get("filters"):
        if (
            "input_field" not in nc_filter_config
            or nc_filter_config["input_field"] is None
        ):
            nc_filter_config["input_field"] = text_field
        new_filter = build_filter(nc_filter_config)
        filters.append(new_filter)

    return nemo_curator.Sequential(filters)


def build_downloader(downloader_config_file, default_download_dir=None):
    # Get the downloader config file
    with open(downloader_config_file, "r") as config_file:
        downloader_params = yaml.load(config_file, Loader=yaml.FullLoader)

    download_class = import_downloader(downloader_params["download_module"])
    no_download_dir = ("download_dir" not in downloader_params["download_params"]) or (
        downloader_params["download_params"] is None
    )
    if no_download_dir and default_download_dir:
        downloader_params["download_params"]["download_dir"] = default_download_dir
    expand_outdir_and_mkdir(downloader_params["download_params"]["download_dir"])
    downloader = download_class(**downloader_params["download_params"])

    iterator_class = import_iterator(downloader_params["iterator_module"])
    iterator = iterator_class(**downloader_params["iterator_params"])

    extractor_class = import_extractor(downloader_params["extract_module"])
    extractor = extractor_class(**downloader_params["extract_params"])

    dataset_format = {}
    for field, field_type in downloader_params["format"].items():
        dataset_format[field] = locate(field_type)

    return downloader, iterator, extractor, dataset_format



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

import dask.dataframe as dd
import pandas as pd
import pytest

import nemo_curator
from nemo_curator.datasets import DocumentDataset
from nemo_curator.tasks import DownstreamTask


class SimpleTask(DownstreamTask):
    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "simple"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = [
            "This is a simple example task document with enough words.",
            "Two plus two equals four. Two times two equals four as well. However, two divided by two is one.",
            "Cars are (sometimes) red. Bicycles can (occasionally) be blue. Trees often have green leaves.",
            "This is a short one.",
            "This is a short one.",
        ]

    def generate_ngrams(self):
        for line in self._dataset:
            self._update_ngrams(line, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class TinyTask(DownstreamTask):
    def __init__(self, min_ngram_size=0, max_ngram_size=4):
        super().__init__()
        self._task_name = "tiny"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = ["Super tiny task with one document."]

    def generate_ngrams(self):
        for line in self._dataset:
            self._update_ngrams(line, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


def list_to_dataset(documents, col_name="text", npartitions=2):
    data = {col_name: documents}
    pdf = pd.DataFrame(data)

    return DocumentDataset(dd.from_pandas(pdf, npartitions=npartitions))


@pytest.fixture
def contaminated_dataset():
    return list_to_dataset(
        [
            "This document is fine",
            "Before contamination. This is a simple example task document with enough words. After contamination.",
            "This document is not good. Two plus two equals four. Two times two equals four as well. However, one minus one is zero.",
            "Long contamination. Birds are (sometimes) red. Bicycles can (occasionally) be blue. Trees often have green leaves. After contamination.",
            "Small contamination in a very short document 1. SupeR tiNy task with one document!",
            "Small contamination in a very short document 2. Super tiNy task with one document?",
        ],
        col_name="text",
    )


class TestPrepareTaskData:
    def test_single_task(self):
        decontaminator = nemo_curator.TaskDecontamination(SimpleTask())
        actual_count = decontaminator.prepare_task_ngram_count().compute()
        expected_ngrams = [
            "this is a simple example task document with enough words",
            "two plus two equals four two times two equals four as well however",
            "plus two equals four two times two equals four as well however two",
            "two equals four two times two equals four as well however two divided",
            "equals four two times two equals four as well however two divided by",
            "four two times two equals four as well however two divided by two",
            "two times two equals four as well however two divided by two is",
            "times two equals four as well however two divided by two is one",
            "cars are sometimes red bicycles can occasionally be blue trees often have green",
            "are sometimes red bicycles can occasionally be blue trees often have green leaves",
        ]
        expected_count = {ngram: 0 for ngram in expected_ngrams}

        assert len(expected_count) == len(
            actual_count
        ), f"Expected #{len(expected_count)}, got #{len(actual_count)}"
        assert (
            expected_count == actual_count
        ), f"Expected: {expected_count}, got: {actual_count}"

    def test_multiple_tasks(self):
        decontaminator = nemo_curator.TaskDecontamination([SimpleTask(), TinyTask()])
        actual_count = decontaminator.prepare_task_ngram_count().compute()
        expected_ngrams = [
            "this is a simple example task document with enough words",
            "two plus two equals four two times two equals four as well however",
            "plus two equals four two times two equals four as well however two",
            "two equals four two times two equals four as well however two divided",
            "equals four two times two equals four as well however two divided by",
            "four two times two equals four as well however two divided by two",
            "two times two equals four as well however two divided by two is",
            "times two equals four as well however two divided by two is one",
            "cars are sometimes red bicycles can occasionally be blue trees often have green",
            "are sometimes red bicycles can occasionally be blue trees often have green leaves",
            "super tiny task with",
            "tiny task with one",
            "task with one document",
        ]
        expected_count = {ngram: 0 for ngram in expected_ngrams}

        assert len(expected_count) == len(
            actual_count
        ), f"Expected #{len(expected_count)}, got #{len(actual_count)}"
        assert (
            expected_count == actual_count
        ), f"Expected: {expected_count}, got: {actual_count}"


class TestFindMatchingNgrams:
    def test_single_task(self, contaminated_dataset):
        decontaminator = nemo_curator.TaskDecontamination(SimpleTask())
        task_ngrams = decontaminator.prepare_task_ngram_count()
        actual_result = decontaminator.find_matching_ngrams(
            task_ngrams, contaminated_dataset
        ).compute()
        actual_ngrams, actual_freq = (
            actual_result["matched-ngrams"],
            actual_result["ngrams-freq"],
        )

        expected_ngrams = defaultdict(int)
        expected_ngrams["this is a simple example task document with enough words"] = 1
        expected_ngrams[
            "two plus two equals four two times two equals four as well however"
        ] = 1
        expected_ngrams[
            "are sometimes red bicycles can occasionally be blue trees often have green leaves"
        ] = 1

        expected_freq = [(10, 1), (13, 9)]

        assert (
            expected_freq == actual_freq
        ), f"Expected #{expected_freq}, got #{actual_freq}"
        assert (
            expected_ngrams == actual_ngrams
        ), f"Expected: {expected_ngrams}, got: {actual_ngrams}"

    def test_multiple_tasks(self, contaminated_dataset):
        decontaminator = nemo_curator.TaskDecontamination([SimpleTask(), TinyTask()])
        task_ngrams = decontaminator.prepare_task_ngram_count()
        actual_result = decontaminator.find_matching_ngrams(
            task_ngrams, contaminated_dataset
        ).compute()
        actual_ngrams, actual_freq = (
            actual_result["matched-ngrams"],
            actual_result["ngrams-freq"],
        )

        expected_ngrams = defaultdict(int)
        expected_ngrams["this is a simple example task document with enough words"] = 1
        expected_ngrams[
            "two plus two equals four two times two equals four as well however"
        ] = 1
        expected_ngrams[
            "are sometimes red bicycles can occasionally be blue trees often have green leaves"
        ] = 1
        expected_ngrams["super tiny task with"] = 2

        expected_freq = [(4, 3), (10, 1), (13, 9)]

        assert (
            expected_freq == actual_freq
        ), f"Expected #{expected_freq}, got #{actual_freq}"
        assert (
            expected_ngrams == actual_ngrams
        ), f"Expected: {expected_ngrams}, got: {actual_ngrams}"


class TestRemoveMatchingNgrams:
    def test_single_task(self, contaminated_dataset):
        decontaminator = nemo_curator.TaskDecontamination(
            SimpleTask(), min_document_length=1, remove_char_each_side=1
        )
        task_ngrams = decontaminator.prepare_task_ngram_count()
        ngram_data = decontaminator.find_matching_ngrams(
            task_ngrams, contaminated_dataset
        )
        matched_ngrams, ngram_freq = (
            ngram_data["matched-ngrams"],
            ngram_data["ngrams-freq"],
        )
        filtered_dataset = decontaminator.remove_matching_ngrams(
            matched_ngrams, ngram_freq, contaminated_dataset
        )

        actual_data = sorted(filtered_dataset.df.compute()["text"].to_list())
        expected_data = [
            "This document is fine",
            "Before contamination.",
            " After contamination.",
            "This document is not good.",
            "Long contamination.",
            " After contamination.",
            "Small contamination in a very short document 1. SupeR tiNy task with one document!",
            "Small contamination in a very short document 2. Super tiNy task with one document?",
        ]
        expected_data.sort()
        assert (
            expected_data == actual_data
        ), f"Expected #{expected_data}, got #{actual_data}"

    def test_multiple_tasks(self, contaminated_dataset):
        decontaminator = nemo_curator.TaskDecontamination(
            [SimpleTask(), TinyTask()], min_document_length=1, remove_char_each_side=1
        )
        task_ngrams = decontaminator.prepare_task_ngram_count()
        ngram_data = decontaminator.find_matching_ngrams(
            task_ngrams, contaminated_dataset
        )
        matched_ngrams, ngram_freq = (
            ngram_data["matched-ngrams"],
            ngram_data["ngrams-freq"],
        )
        filtered_dataset = decontaminator.remove_matching_ngrams(
            matched_ngrams, ngram_freq, contaminated_dataset
        )

        actual_data = sorted(filtered_dataset.df.compute()["text"].to_list())
        expected_data = [
            "This document is fine",
            "Before contamination.",
            " After contamination.",
            "This document is not good.",
            "Long contamination.",
            " After contamination.",
            "Small contamination in a very short document 1.",
            "Small contamination in a very short document 2.",
        ]
        expected_data.sort()
        assert (
            expected_data == actual_data
        ), f"Expected #{expected_data}, got #{actual_data}"


class TestFullPipeline:
    def test_single_task(self, contaminated_dataset):
        decontaminator = nemo_curator.TaskDecontamination(
            SimpleTask(), min_document_length=1, remove_char_each_side=1
        )
        filtered_dataset = decontaminator(contaminated_dataset)

        actual_data = sorted(filtered_dataset.df.compute()["text"].to_list())
        expected_data = [
            "This document is fine",
            "Before contamination.",
            " After contamination.",
            "This document is not good.",
            "Long contamination.",
            " After contamination.",
            "Small contamination in a very short document 1. SupeR tiNy task with one document!",
            "Small contamination in a very short document 2. Super tiNy task with one document?",
        ]
        expected_data.sort()
        assert (
            expected_data == actual_data
        ), f"Expected #{expected_data}, got #{actual_data}"

    def test_multiple_tasks(self, contaminated_dataset):
        decontaminator = nemo_curator.TaskDecontamination(
            [SimpleTask(), TinyTask()], min_document_length=1, remove_char_each_side=1
        )
        filtered_dataset = decontaminator(contaminated_dataset)

        actual_data = sorted(filtered_dataset.df.compute()["text"].to_list())
        expected_data = [
            "This document is fine",
            "Before contamination.",
            " After contamination.",
            "This document is not good.",
            "Long contamination.",
            " After contamination.",
            "Small contamination in a very short document 1.",
            "Small contamination in a very short document 2.",
        ]
        expected_data.sort()
        assert (
            expected_data == actual_data
        ), f"Expected #{expected_data}, got #{actual_data}"



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import time

from nemo_curator import MinHash
from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import (
    get_client,
    performance_report_if,
    read_data,
)
from nemo_curator.utils.file_utils import get_all_files_paths_under
from nemo_curator.utils.fuzzy_dedup_utils.io_utils import strip_trailing_sep
from nemo_curator.utils.script_utils import ArgumentHelper


def pre_imports():
    import cudf  # noqa: F401


def main(args):
    logger = create_logger(
        rank=0, log_file=os.path.join(args.log_dir, "rank_000.log"), name="minhash_log"
    )
    logger.info(f"Starting workflow with args:\n {args}")

    assert args.hash_bytes in {4, 8}, "Currently only 32bit/64bit hashes are supported"
    assert args.device == "gpu"

    client = get_client(**ArgumentHelper.parse_client_args(args))
    logger.info(f"Client Created {client}")
    client.run(pre_imports)
    logger.info("Pre imports complete")

    data_paths = args.input_data_dirs
    id_field = args.input_json_id_field
    text_field = args.input_json_text_field
    num_files = args.num_files

    minhasher = MinHash(
        seed=args.seed,
        num_hashes=args.minhash_length,
        char_ngrams=args.char_ngram,
        use_64bit_hash=False if args.hash_bytes == 4 else True,
        logger=logger,
        id_field=id_field,
        text_field=text_field,
    )

    t0 = time.time()
    for data_path in data_paths:
        print(f"Computing minhashes for {data_path}", flush=True)
        data_path = strip_trailing_sep(data_path)
        if num_files is not None and num_files <= 0:
            print(f"Processed {args.num_files}... quitting")
            break

        files = get_all_files_paths_under(
            root=data_path, recurse_subdirectories=False, keep_extensions="jsonl"
        )
        df = read_data(
            files[:num_files] if num_files else files,
            file_type="jsonl",
            backend="cudf",
            files_per_partition=args.files_per_partition,
            add_filename=False,
            input_meta=args.input_meta,
        )[[id_field, text_field]]

        if num_files is not None:
            num_files -= len(files)

        res = minhasher(DocumentDataset(df)).df
        logger.info(
            f"Lazy minhash generation complete for {res.npartitions} partitions"
        )
        logger.info(f"Starting execution for {data_path}")
        write_path = os.path.join(
            args.output_minhash_dir, os.path.basename(data_path), "minhashes.parquet"
        )

        t1 = time.time()
        with performance_report_if(
            args.profile_path, f"{os.path.basename(data_path)}-minhash-profile.html"
        ):
            res.to_parquet(write_path, write_index=False)
        logger.info(
            f"Minhash computation for f{data_path} took {time.time() - t1}s complete at {write_path}"  # noqa:E501
        )
    logger.info(
        f"Minhash computation across datasets took {time.time() - t0}s complete at {args.output_minhash_dir}"  # noqa:E501
    )


def attach_args():
    description = """
    Computes minhash signatures from an input directory of documents
    contained within JSONL files. For each document, a DataFrame of document IDs
    and minhash signatures is created. This DataFrame is written to file after processing.
    """
    parser = argparse.ArgumentParser(
        description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.parse_gpu_dedup_args()

    argumentHelper.add_arg_minhash_length()
    argumentHelper.add_arg_seed(
        help="Random seed used for intializing the hash "
        "functions used to compute the minhashes."
    )
    argumentHelper.add_arg_input_meta()
    parser.add_argument(
        "--char-ngram",
        type=int,
        default=24,
        help="The number of consecutive characters to include in a sliding "
        "window when creating the document shingles for computing "
        "minhash signatures.",
    )
    parser.add_argument(
        "--hash-bytes",
        type=int,
        default=4,
        help="Number of bytes per computed minhash. "
        "Default is an unsigned 32-bit integer.",
    )
    parser.add_argument(
        "--output-minhash-dir",
        type=str,
        required=True,
        help="Output directory where minhashes will be written. "
        "Each file is a Parquet file that contains two series, the document IDs, "
        "and a series of lists, each list denoting the minhash signature for that document ID.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



import pandas as pd

from nemo_curator.datasets import DocumentDataset


def test_to_from_pandas():
    original_df = pd.DataFrame({"first_col": [1, 2, 3], "second_col": ["a", "b", "c"]})
    dataset = DocumentDataset.from_pandas(original_df)
    converted_df = dataset.to_pandas()
    pd.testing.assert_frame_equal(original_df, converted_df)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

GPU_INSTALL_STRING = """Install GPU packages via `pip install --extra-index-url https://pypi.nvidia.com nemo-curator[cuda12x]`
or use `pip install --extra-index-url https://pypi.nvidia.com ".[cuda12x]"` if installing from source"""


def is_cudf_type(obj):
    """
    Check if an object is a cuDF type
    """
    types = [
        str(type(obj)),
        str(getattr(obj, "_partition_type", "")),
        str(getattr(obj, "_meta", "")),
    ]
    return any("cudf" in obj_type for obj_type in types)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dask.dataframe
import pandas as pd

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modifiers.llm_pii_modifier import LLMPiiModifier
from nemo_curator.modules.modify import Modify
from nemo_curator.utils.distributed_utils import get_client


def console_script():
    _ = get_client()

    dataframe = pd.DataFrame(
        {
            "text": [
                # Sampled from https://huggingface.co/datasets/gretelai/gretel-pii-masking-en-v1
                "Transaction details: gasLimit set to 1000000 units by tw_brian740, gasPrice set to 10 Gwei by veronicawood@example.org, contactable at +1-869-341-9301x7005, located at Suite 378, Yolanda Mountain, Burkeberg.",
                "Unloading Plan for Shipment MRN-293104, MED25315002, dated 1989.12.22. Driver EMP730359, Vehicle KS40540825.",
            ]
        }
    )
    dataset = DocumentDataset.from_pandas(dataframe, npartitions=1)

    modifier = LLMPiiModifier(
        # Endpoint for the user's NIM
        base_url="http://0.0.0.0:8000/v1",
        api_key="API KEY (if needed)",
        model="meta/llama-3.1-70b-instruct",
    )

    modify = Modify(modifier)
    modified_dataset = modify(dataset)
    modified_dataset.to_json("output.jsonl")


if __name__ == "__main__":
    console_script()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import tempfile
from unittest import mock

import pytest
from datasets import Dataset

import nemo_curator
from nemo_curator.tasks.metrics import (
    ANLI,
    CB,
    COQA,
    MMLU,
    PIQA,
    RTE,
    WSC,
    ArcChallenge,
    ArcEasy,
    BigBenchHard,
    BigBenchLight,
    BoolQ,
    Copa,
    Drop,
    Lambada,
    Multilingual,
    MultiRC,
    NumDasc,
    OpenBookQA,
    Quac,
    Race,
    Record,
    Squad,
    StoryCloze,
    TriviaQA,
    WebQA,
    WiC,
    Winogrande,
)


# Mock dataset for testing
@pytest.fixture
def mock_dataset():
    return [
        {"question": "This is a test question for metrics testing"},
        {"question": "Here is another question with more words for ngram generation"},
        {"question": "Very short"},
        {
            "question": "This contains enough words to create multiple different ngrams for testing purposes"
        },
    ]


@pytest.fixture
def mock_question_stem_dataset():
    return [
        {"question_stem": "This is a test question stem for metrics testing"},
        {
            "question_stem": "Here is another question stem with more words for ngram generation"
        },
        {"question_stem": "Very short stem"},
        {
            "question_stem": "This question stem contains enough words to create multiple different ngrams for testing purposes"
        },
    ]


@pytest.fixture
def mock_premises_dataset():
    return [
        {"premise": "This is a test premise for metrics testing"},
        {"premise": "Here is another premise with more words for ngram generation"},
        {"premise": "Very short premise"},
        {
            "premise": "This premise contains enough words to create multiple different ngrams for testing purposes"
        },
    ]


@pytest.fixture
def mock_sentences_dataset():
    return [
        {
            "sentence1": "This is the first test sentence",
            "sentence2": "This is the second test sentence",
        },
        {
            "sentence1": "Here is another long first sentence with words",
            "sentence2": "And a matching second sentence",
        },
        {"sentence1": "Short one", "sentence2": "Another short"},
        {"sentence1": "This sentence contains", "sentence2": "enough words for ngrams"},
    ]


@pytest.fixture
def mock_coqa_data():
    return {
        "data": [
            {
                "questions": [
                    {"input_text": "This is a test question for COQA?"},
                    {
                        "input_text": "Another question with more words for testing COQA?"
                    },
                ],
                "story": "This is a test story for COQA that has enough words for ngram generation.",
            },
            {
                "questions": [{"input_text": "Short question?"}],
                "story": "Another story with sufficient length for ngram creation.",
            },
        ]
    }


@pytest.fixture
def mock_jsonl_data():
    return [
        '{"text": "This is a test document for MMLU with enough words to create ngrams."}\n',
        '{"text": "Here is another line for BigBench testing with sufficient words."}\n',
        '{"text": "More text for testing multilingual capabilities with proper length."}\n',
        '{"text": "Short."}\n',
    ]


@pytest.fixture
def mock_lambada_data():
    return [
        '{"text": "This is a test document for Lambada with enough words to create ngrams."}\n',
        '{"text": "Here is another line for Lambada testing with sufficient words."}\n',
        '{"text": "Short."}\n',
    ]


@pytest.fixture
def mock_numdasc_data():
    return [
        '{"context": "This is the context", "completion": "with the completion part for testing."}\n',
        '{"context": "Another context section", "completion": "with its completion that should have sufficient words for ngrams."}\n',
        '{"context": "Short", "completion": "one."}\n',
    ]


@pytest.fixture
def mock_storycloze_data():
    return [
        '{"InputSentence1": "First sentence of story.", "InputSentence2": "Second part continues.", "InputSentence3": "Third section adds detail.", "InputSentence4": "Fourth concludes test story."}\n',
        '{"InputSentence1": "Story begins here.", "InputSentence2": "It continues with details.", "InputSentence3": "Adding more narrative.", "InputSentence4": "Concluding the test story."}\n',
    ]


@pytest.fixture
def mock_query_dataset():
    return [
        {"query": "This is a test query for Record testing"},
        {"query": "Here is another query with more words for ngram generation"},
        {"query": "Very short query"},
        {
            "query": "This query contains enough words to create multiple different ngrams for testing purposes"
        },
    ]


class TestMetricsTasks:

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_race_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        # Setup mock
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        # Test initialization
        race_task = Race(min_ngram_size=5, max_ngram_size=10)
        assert race_task._task_name == "race"
        assert race_task._min_ngram_size == 5
        assert race_task._max_ngram_size == 10

        # Test generate_ngrams
        ngrams = race_task.generate_ngrams()
        assert len(ngrams) > 0

        # Verify ngram content
        # The exact count depends on the mock dataset but we can check if expected ngrams are present
        expected_ngram = "this is a test question for metrics"
        assert any(expected_ngram in ngram for ngram in ngrams.keys())

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_squad_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        # Setup mock
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        # Test initialization
        squad_task = Squad(min_ngram_size=5, max_ngram_size=10)
        assert squad_task._task_name == "squad"
        assert squad_task._min_ngram_size == 5
        assert squad_task._max_ngram_size == 10

        # Test generate_ngrams
        ngrams = squad_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_arceasy_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        arceasy_task = ArcEasy(min_ngram_size=5, max_ngram_size=10)
        assert arceasy_task._task_name == "arceasy"
        assert arceasy_task._min_ngram_size == 5
        assert arceasy_task._max_ngram_size == 10

        ngrams = arceasy_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_arcchallenge_init_and_generate_ngrams(
        self, mock_load_dataset, mock_dataset
    ):
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        arcchallenge_task = ArcChallenge(min_ngram_size=5, max_ngram_size=10)
        assert arcchallenge_task._task_name == "arcchallenge"
        assert arcchallenge_task._min_ngram_size == 5
        assert arcchallenge_task._max_ngram_size == 10

        ngrams = arcchallenge_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_openbookqa_init_and_generate_ngrams(
        self, mock_load_dataset, mock_question_stem_dataset
    ):
        mock_load_dataset.return_value = Dataset.from_list(mock_question_stem_dataset)

        openbookqa_task = OpenBookQA(min_ngram_size=5, max_ngram_size=10)
        assert openbookqa_task._task_name == "openbookqa"
        assert openbookqa_task._min_ngram_size == 5
        assert openbookqa_task._max_ngram_size == 10

        ngrams = openbookqa_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_boolq_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        boolq_task = BoolQ(min_ngram_size=5, max_ngram_size=10)
        assert boolq_task._task_name == "boolq"
        assert boolq_task._min_ngram_size == 5
        assert boolq_task._max_ngram_size == 10

        ngrams = boolq_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_copa_init_and_generate_ngrams(
        self, mock_load_dataset, mock_premises_dataset
    ):
        mock_load_dataset.return_value = Dataset.from_list(mock_premises_dataset)

        copa_task = Copa(min_ngram_size=5, max_ngram_size=10)
        assert copa_task._task_name == "copa"
        assert copa_task._min_ngram_size == 5
        assert copa_task._max_ngram_size == 10

        ngrams = copa_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_rte_init_and_generate_ngrams(
        self, mock_load_dataset, mock_sentences_dataset
    ):
        mock_load_dataset.return_value = Dataset.from_list(mock_sentences_dataset)

        rte_task = RTE(min_ngram_size=5, max_ngram_size=10)
        assert rte_task._task_name == "rte"
        assert rte_task._min_ngram_size == 5
        assert rte_task._max_ngram_size == 10

        ngrams = rte_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_multirc_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        multirc_task = MultiRC(min_ngram_size=5, max_ngram_size=10)
        assert multirc_task._task_name == "multirc"
        assert multirc_task._min_ngram_size == 5
        assert multirc_task._max_ngram_size == 10

        ngrams = multirc_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_wsc_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        wsc_task = WSC(min_ngram_size=5, max_ngram_size=10)
        assert wsc_task._task_name == "wsc"
        assert wsc_task._min_ngram_size == 5
        assert wsc_task._max_ngram_size == 10

        ngrams = wsc_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_cb_init_and_generate_ngrams(
        self, mock_load_dataset, mock_premises_dataset
    ):
        mock_load_dataset.return_value = Dataset.from_list(mock_premises_dataset)

        cb_task = CB(min_ngram_size=5, max_ngram_size=10)
        assert cb_task._task_name == "cb"
        assert cb_task._min_ngram_size == 5
        assert cb_task._max_ngram_size == 10

        ngrams = cb_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_anli_init_and_generate_ngrams(
        self, mock_load_dataset, mock_premises_dataset
    ):
        mock_dataset_dict = {
            "test_r1": Dataset.from_list(mock_premises_dataset),
            "test_r2": Dataset.from_list(mock_premises_dataset),
            "test_r3": Dataset.from_list(mock_premises_dataset),
        }
        mock_load_dataset.return_value = mock_dataset_dict

        anli_task = ANLI(min_ngram_size=5, max_ngram_size=10)
        assert anli_task._task_name == "anli"
        assert anli_task._min_ngram_size == 5
        assert anli_task._max_ngram_size == 10

        ngrams = anli_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_record_init_and_generate_ngrams(
        self, mock_load_dataset, mock_query_dataset
    ):
        mock_load_dataset.return_value = Dataset.from_list(mock_query_dataset)

        record_task = Record(min_ngram_size=5, max_ngram_size=10)
        assert record_task._task_name == "record"
        assert record_task._min_ngram_size == 5
        assert record_task._max_ngram_size == 10

        ngrams = record_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_coqa_init_and_generate_ngrams(self, mock_coqa_data, tmp_path):
        # Create a temporary file with mock COQA data
        coqa_file = tmp_path / "coqa.json"
        with open(coqa_file, "w") as f:
            json.dump(mock_coqa_data, f)

        coqa_task = COQA(file_path=str(coqa_file), min_ngram_size=5, max_ngram_size=10)
        assert coqa_task._task_name == "coqa"
        assert coqa_task._min_ngram_size == 5
        assert coqa_task._max_ngram_size == 10

        ngrams = coqa_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_coqa_raises_exception_without_file_path(self):
        with pytest.raises(
            Exception, match="Must provide a path to the coqa.json file"
        ):
            COQA(file_path=None)

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_triviaqa_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        triviaqa_task = TriviaQA(min_ngram_size=5, max_ngram_size=10)
        assert triviaqa_task._task_name == "trivia_qa"
        assert triviaqa_task._min_ngram_size == 5
        assert triviaqa_task._max_ngram_size == 10

        ngrams = triviaqa_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_quac_init_and_generate_ngrams(self, mock_load_dataset):
        # For Quac, setup a specific dataset structure
        mock_dataset_list = [
            {
                "questions": [
                    "Question one for quac?",
                    "Question two with more words for quac testing?",
                ]
            },
            {
                "questions": [
                    "Another quac question with sufficient words for ngram generation?"
                ]
            },
        ]
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset_list)

        quac_task = Quac(min_ngram_size=5, max_ngram_size=10)
        assert quac_task._task_name == "quac"
        assert quac_task._min_ngram_size == 5
        assert quac_task._max_ngram_size == 10

        ngrams = quac_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_webqa_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        webqa_task = WebQA(min_ngram_size=5, max_ngram_size=10)
        assert webqa_task._task_name == "webqa"
        assert webqa_task._min_ngram_size == 5
        assert webqa_task._max_ngram_size == 10

        ngrams = webqa_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_drop_init_and_generate_ngrams(self, mock_load_dataset, mock_dataset):
        mock_load_dataset.return_value = Dataset.from_list(mock_dataset)

        drop_task = Drop(min_ngram_size=5, max_ngram_size=10)
        assert drop_task._task_name == "drop"
        assert drop_task._min_ngram_size == 5
        assert drop_task._max_ngram_size == 10

        ngrams = drop_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_wic_init_and_generate_ngrams(
        self, mock_load_dataset, mock_sentences_dataset
    ):
        mock_load_dataset.return_value = Dataset.from_list(mock_sentences_dataset)

        wic_task = WiC(min_ngram_size=5, max_ngram_size=10)
        assert wic_task._task_name == "wic"
        assert wic_task._min_ngram_size == 5
        assert wic_task._max_ngram_size == 10

        ngrams = wic_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.utils.file_utils.get_all_files_paths_under")
    def test_mmlu_init_and_generate_ngrams(
        self, mock_get_paths, mock_jsonl_data, tmp_path
    ):
        # Create a temporary file with mock JSONL data
        mmlu_dir = tmp_path / "mmlu"
        mmlu_dir.mkdir()
        mmlu_file = mmlu_dir / "data.jsonl"
        with open(mmlu_file, "w") as f:
            f.writelines(mock_jsonl_data)

        mock_get_paths.return_value = [str(mmlu_file)]

        mmlu_task = MMLU(path=str(mmlu_dir), min_ngram_size=5, max_ngram_size=10)
        assert mmlu_task._task_name == "mmlu"
        assert mmlu_task._min_ngram_size == 5
        assert mmlu_task._max_ngram_size == 10

        ngrams = mmlu_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_mmlu_raises_exception_without_path(self):
        with pytest.raises(Exception, match="Must provide path that contain"):
            MMLU(path=None)

    @mock.patch("nemo_curator.utils.file_utils.get_all_files_paths_under")
    def test_bigbenchhard_init_and_generate_ngrams(
        self, mock_get_paths, mock_jsonl_data, tmp_path
    ):
        bbh_dir = tmp_path / "bbh"
        bbh_dir.mkdir()
        bbh_file = bbh_dir / "data.jsonl"
        with open(bbh_file, "w") as f:
            f.writelines(mock_jsonl_data)

        mock_get_paths.return_value = [str(bbh_file)]

        bbh_task = BigBenchHard(path=str(bbh_dir), min_ngram_size=5, max_ngram_size=10)
        assert bbh_task._task_name == "bigbench_hard"
        assert bbh_task._min_ngram_size == 5
        assert bbh_task._max_ngram_size == 10

        ngrams = bbh_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_bigbenchhard_raises_exception_without_path(self):
        with pytest.raises(Exception, match="Must provide path that contain"):
            BigBenchHard(path=None)

    @mock.patch("nemo_curator.utils.file_utils.get_all_files_paths_under")
    def test_bigbenchlight_init_and_generate_ngrams(
        self, mock_get_paths, mock_jsonl_data, tmp_path
    ):
        bbl_dir = tmp_path / "bbl"
        bbl_dir.mkdir()
        bbl_file = bbl_dir / "data.jsonl"
        with open(bbl_file, "w") as f:
            f.writelines(mock_jsonl_data)

        mock_get_paths.return_value = [str(bbl_file)]

        bbl_task = BigBenchLight(path=str(bbl_dir), min_ngram_size=5, max_ngram_size=10)
        assert bbl_task._task_name == "bigbench_light"
        assert bbl_task._min_ngram_size == 5
        assert bbl_task._max_ngram_size == 10

        ngrams = bbl_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_bigbenchlight_raises_exception_without_path(self):
        with pytest.raises(Exception, match="Must provide path that contain"):
            BigBenchLight(path=None)

    @mock.patch("nemo_curator.utils.file_utils.get_all_files_paths_under")
    def test_multilingual_init_and_generate_ngrams(
        self, mock_get_paths, mock_jsonl_data, tmp_path
    ):
        ml_dir = tmp_path / "multilingual"
        ml_dir.mkdir()
        ml_file = ml_dir / "data.jsonl"
        with open(ml_file, "w") as f:
            f.writelines(mock_jsonl_data)

        mock_get_paths.return_value = [str(ml_file)]

        ml_task = Multilingual(path=str(ml_dir), min_ngram_size=5, max_ngram_size=10)
        assert ml_task._task_name == "multilingual"
        assert ml_task._min_ngram_size == 5
        assert ml_task._max_ngram_size == 10

        ngrams = ml_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_multilingual_raises_exception_without_path(self):
        with pytest.raises(Exception, match="Must provide path to"):
            Multilingual(path=None)

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_piqa_init_and_generate_ngrams(self, mock_load_dataset):
        # For PIQA, setup a specific dataset structure
        mock_piqa_dataset = [
            {
                "goal": "This is a goal for PIQA with enough words to test ngram generation."
            },
            {"goal": "Another goal with sufficient length for PIQA testing."},
            {"goal": "Short goal."},
        ]
        mock_load_dataset.return_value = Dataset.from_list(mock_piqa_dataset)

        piqa_task = PIQA(min_ngram_size=5, max_ngram_size=10)
        assert piqa_task._task_name == "piqa"
        assert piqa_task._min_ngram_size == 5
        assert piqa_task._max_ngram_size == 10

        ngrams = piqa_task.generate_ngrams()
        assert len(ngrams) > 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_winogrande_init_and_generate_ngrams(self, mock_load_dataset):
        # For Winogrande, setup a specific dataset structure
        mock_winogrande_dataset = [
            {
                "sentence": "This is a Winogrande sentence with enough words to test ngram generation."
            },
            {
                "sentence": "Another sentence with sufficient length for Winogrande testing."
            },
            {"sentence": "Short sentence."},
        ]
        mock_load_dataset.return_value = Dataset.from_list(mock_winogrande_dataset)

        winogrande_task = Winogrande(min_ngram_size=5, max_ngram_size=10)
        assert winogrande_task._task_name == "winogrande"
        assert winogrande_task._min_ngram_size == 5
        assert winogrande_task._max_ngram_size == 10

        ngrams = winogrande_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_lambada_init_and_generate_ngrams(self, mock_lambada_data, tmp_path):
        # Create a temporary file with mock Lambada data
        lambada_file = tmp_path / "lambada.jsonl"
        with open(lambada_file, "w") as f:
            f.writelines(mock_lambada_data)

        lambada_task = Lambada(
            file_path=str(lambada_file), min_ngram_size=5, max_ngram_size=10
        )
        assert lambada_task._task_name == "lambada"
        assert lambada_task._min_ngram_size == 5
        assert lambada_task._max_ngram_size == 10

        ngrams = lambada_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_numdadc_init_and_generate_ngrams(self, mock_numdasc_data, tmp_path):
        # Create a temporary file with mock NumDasc data
        numdasc_file = tmp_path / "numdasc.jsonl"
        with open(numdasc_file, "w") as f:
            f.writelines(mock_numdasc_data)

        numdasc_task = NumDasc(
            n=2, file_path=str(numdasc_file), min_ngram_size=5, max_ngram_size=10
        )
        assert numdasc_task._task_name == "{n}dasc"
        assert numdasc_task._min_ngram_size == 5
        assert numdasc_task._max_ngram_size == 10
        assert numdasc_task._n == 2

        ngrams = numdasc_task.generate_ngrams()
        assert len(ngrams) > 0

    def test_storycloze_init_and_generate_ngrams(self, mock_storycloze_data, tmp_path):
        # Create a temporary file with mock StoryCloze data
        storycloze_file = tmp_path / "storycloze.jsonl"
        with open(storycloze_file, "w") as f:
            f.writelines(mock_storycloze_data)

        storycloze_task = StoryCloze(
            file_path=str(storycloze_file), min_ngram_size=5, max_ngram_size=10
        )
        assert storycloze_task._task_name == "story_cloze"
        assert storycloze_task._min_ngram_size == 5
        assert storycloze_task._max_ngram_size == 10

        ngrams = storycloze_task.generate_ngrams()
        assert len(ngrams) > 0

    # Tests for exception handling in Lambada, NumDasc, and StoryCloze

    def test_lambada_exception_handling(self, tmp_path):
        """Test that Lambada properly handles invalid JSON and prints error messages."""
        # Create a file with invalid JSON data
        invalid_file = tmp_path / "invalid_lambada.jsonl"
        with open(invalid_file, "w") as f:
            f.write('{"text": "Valid JSON line"}\n')
            f.write("Invalid JSON line that will cause an exception\n")
            f.write('{"incomplete": "Missing text field"}\n')

        lambada_task = Lambada(
            file_path=str(invalid_file), min_ngram_size=5, max_ngram_size=10
        )

        # Mock stdout to capture printed error messages
        with mock.patch("builtins.print") as mock_print:
            # This should not raise an exception
            ngrams = lambada_task.generate_ngrams()

            # Verify error was printed
            mock_print.assert_called()
            # At least one call should have "Error" in it
            assert any("Error" in str(call) for call in mock_print.call_args_list)

    def test_numdasc_exception_handling(self, tmp_path):
        """Test that NumDasc properly handles invalid JSON and prints error messages."""
        # Create a file with invalid JSON data
        invalid_file = tmp_path / "invalid_numdasc.jsonl"
        with open(invalid_file, "w") as f:
            f.write('{"context": "Valid", "completion": "JSON line"}\n')
            f.write("Invalid JSON line\n")
            f.write('{"context": "Missing completion field"}\n')

        numdasc_task = NumDasc(
            n=2, file_path=str(invalid_file), min_ngram_size=5, max_ngram_size=10
        )

        # Mock stdout to capture printed error messages
        with mock.patch("builtins.print") as mock_print:
            # This should not raise an exception
            ngrams = numdasc_task.generate_ngrams()

            # Verify error was printed
            mock_print.assert_called()
            # At least one call should have "Error" in it
            assert any("Error" in str(call) for call in mock_print.call_args_list)

    def test_storycloze_exception_handling(self, tmp_path):
        """Test that StoryCloze properly handles invalid JSON and prints error messages."""
        # Create a file with invalid JSON data
        invalid_file = tmp_path / "invalid_storycloze.jsonl"
        with open(invalid_file, "w") as f:
            f.write(
                '{"InputSentence1": "Valid", "InputSentence2": "JSON", "InputSentence3": "line", "InputSentence4": "here"}\n'
            )
            f.write("Invalid JSON line\n")
            f.write('{"InputSentence1": "Missing other required fields"}\n')

        storycloze_task = StoryCloze(
            file_path=str(invalid_file), min_ngram_size=5, max_ngram_size=10
        )

        # Mock stdout to capture printed error messages
        with mock.patch("builtins.print") as mock_print:
            # This should not raise an exception
            ngrams = storycloze_task.generate_ngrams()

            # Verify error was printed
            mock_print.assert_called()
            # At least one call should have "Error" in it
            assert any("Error" in str(call) for call in mock_print.call_args_list)

    # Edge cases and error handling tests

    def test_task_with_short_texts(self, tmp_path):
        """Test that tasks handle texts shorter than min_ngram_size properly."""
        # Create a file with very short texts
        short_texts_file = tmp_path / "short.jsonl"
        with open(short_texts_file, "w") as f:
            f.write('{"text": "Short."}\n')
            f.write('{"text": "Too short."}\n')

        # Mock the file paths function
        with mock.patch(
            "nemo_curator.utils.file_utils.get_all_files_paths_under"
        ) as mock_get_paths:
            mock_get_paths.return_value = [str(short_texts_file)]

            # Test with a high min_ngram_size
            task = MMLU(path=str(tmp_path), min_ngram_size=20, max_ngram_size=25)
            ngrams = task.generate_ngrams()

            # Should not generate any ngrams from short texts
            assert len(ngrams) == 0

    @mock.patch("nemo_curator.tasks.metrics.load_dataset")
    def test_error_handling_in_anli(self, mock_load_dataset, mock_premises_dataset):
        """Test that ANLI handles errors during processing."""
        # Create a dataset with bad entries that will cause errors
        bad_dataset = mock_premises_dataset.copy()
        # Add an entry without required fields
        bad_dataset.append({"not_premise": "This will cause an error"})

        mock_dataset_dict = {
            "test_r1": Dataset.from_list(bad_dataset),
            "test_r2": Dataset.from_list(mock_premises_dataset),
            "test_r3": Dataset.from_list(mock_premises_dataset),
        }
        mock_load_dataset.return_value = mock_dataset_dict

        # Should not raise an exception despite the bad data
        anli_task = ANLI(min_ngram_size=5, max_ngram_size=10)
        ngrams = anli_task.generate_ngrams()

        # Should still generate ngrams from the valid data
        assert len(ngrams) > 0



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from dataclasses import dataclass
from typing import Literal, Optional, Union

import yaml


@dataclass
class BaseConfig:
    @classmethod
    def from_yaml(cls, file_path: str):
        with open(file_path, "r") as file:
            yaml_dict = yaml.safe_load(file)
        return cls(**yaml_dict)


@dataclass
class FuzzyDuplicatesConfig(BaseConfig):
    """
    Configuration for MinHash based fuzzy duplicates detection.
    Parameters
    ----------
    seed: Seed for minhash permutations
    char_ngrams: Size of Char ngram shingles used in minhash computation
    num_buckets: Number of Bands or buckets to use during Locality Sensitive Hashing
    hashes_per_bucket: Number of hashes per bucket/band.
    use_64_bit_hash: Whether to use a 32bit or 64bit hash function for minhashing.
    buckets_per_shuffle: Number of bands/buckets to shuffle concurrently.
        Larger values process larger batches by processing multiple bands
        but might lead to memory pressures and related errors.
    id_field: Column in the Dataset denoting document ID.
    text_field: Column in the Dataset denoting document content.
    perform_removal: Boolean value to specify whether calling the module should remove the duplicates from
        the original dataset, or return the list of IDs denoting duplicates.
    profile_dir: str, Default None
        If specified directory to write dask profile
    cache_dir: str, Default None
        Location to store deduplcation intermediates such as minhashes/buckets etc.
    false_positive_check: bool,
        Whether to run a check to look for false positives within buckets.
        Note: This is a computationally expensive step.
    num_anchors: int
        Number of documents per bucket to use as reference for computing jaccard
        pairs within that bucket to identify false positives.
    jaccard_threshold: float
        The Jaccard similariy threshold to consider a document a near duplicate
        during false positive evaluations.
    """

    # General config
    cache_dir: str
    profile_dir: Optional[str] = None
    id_field: str = "id"
    text_field: str = "text"
    perform_removal: bool = False

    # Minhash + LSH Config
    seed: int = 42
    char_ngrams: int = 24
    num_buckets: int = 20
    hashes_per_bucket: int = 13
    use_64_bit_hash: bool = False
    buckets_per_shuffle: int = 1

    false_positive_check: bool = False
    # Only required for false positive check
    num_anchors: Optional[int] = None
    jaccard_threshold: Optional[float] = None
    bucket_mapping_blocksize: Optional[int] = None
    parts_per_worker: Optional[int] = None
    bucket_parts_per_worker: Optional[int] = None

    def __post_init__(self):
        self.num_hashes = self.num_buckets * self.hashes_per_bucket
        false_positive_defaults = {
            "num_anchors": 2,
            "jaccard_threshold": 0.8,
            "bucket_mapping_blocksize": 256,
            "parts_per_worker": 1,
            "bucket_parts_per_worker": 8,
        }
        if self.false_positive_check:
            warnings.warn(
                "Identifying false positives during the Minhash deduplication is computationally expensive."
                " For improved performance consider setting this to False"
            )
            for arg, default in false_positive_defaults.items():
                if getattr(self, arg) is None:
                    setattr(self, arg, default)
            if self.num_anchors <= 0:
                raise ValueError("Number of anchors must be greater than 0")
            if self.num_anchors > 2:
                warnings.warn(
                    "Using a higher number of anchor docs might lead to higher memory footprint and might impact performance",
                    category=UserWarning,
                )
            if not 0 <= self.jaccard_threshold <= 1:
                raise ValueError("Jaccard Threshold must be between [0,1]")
        else:
            if self.char_ngrams < 20:
                warnings.warn(
                    "Using a small char_ngrams value might lead to a large number (~5%) of false positives during deduplication."
                    " Using a value of at least 20 for char_ngrams is recommended."
                )
            unused_false_positive_args = [
                arg
                for arg in false_positive_defaults.keys()
                if getattr(self, arg) is not None
            ]
            if unused_false_positive_args:
                warnings.warn(
                    f"False positive check is disabled. Unused arguments {unused_false_positive_args} will be ignored",
                    category=UserWarning,
                )

        if self.cache_dir is None:
            raise ValueError(
                "Finding fuzzy duplicates requires a cache directory accessible via all workers to store intermediates"
            )
        if not 1 <= self.buckets_per_shuffle <= self.num_buckets:
            raise ValueError("Buckets per shuffle must be between [1, num_buckets]")

        if not self.perform_removal:
            warnings.warn(
                "In future NeMo Curator releases, the default value for perform_removal will be True."
            )


@dataclass
class SemDedupConfig(BaseConfig):
    """
    Configuration for Semantic Deduplication.

    Attributes:
        cache_dir (str): Directory to store cache.
        profile_dir (Optional[str]): If specified, directory to write Dask profile.
            Default is None.
        num_files (int): Number of files. Default is -1, meaning all files.

        embedding_model_name_or_path (str): Model name or path for embeddings.
            Default is "sentence-transformers/all-MiniLM-L6-v2".
        embedding_batch_size (int): Initial batch size for processing embeddings.
            Default is 128.
        embeddings_save_loc (str): Location to save embeddings.
            Default is "embeddings".
        embedding_max_mem_gb (int): Maximum memory usage in GB for the embedding process.
            If None, it defaults to the available GPU memory minus 4 GB.
        embedding_pooling_strategy (str): Strategy for pooling embeddings, either
            "mean_pooling" or "last_token". Default is "mean_pooling".
        embedding_column (str): The column name that stores the embeddings.
            Default is "embeddings".
        write_embeddings_to_disk (bool): If True, saves the embeddings to disk.
            We recommend setting this to False when you have a delayed pipeline.
            Setting it to False can lead to more memory overhead. Default is True.
        write_to_filename (bool): If True, saves the embeddings to the same filename as input files.
            Default False.

        max_iter (int): Maximum iterations for clustering. The more iterations, the better the clustering.
            Default is 100.
        n_clusters (int): Number of clusters. Default is 1000.
        clustering_save_loc (str): Location to save clustering results.
            Default is "clustering_results".
        random_state (int): KMeans random state used for reproducibility. Default is 1234.
        sim_metric ("cosine" or "l2"): Similarity metric to use to rank within cluster. Default is "cosine".
            `which_to_keep` determines how points within each cluster are ranked, based on the similarity to the centroid defined by `sim_metric`
        which_to_keep (str): Method to determine which duplicates to keep. Default is "hard".
            - hard retains edge-case or outlier items farthest from the centroid by sorting points by decreasing distance from the centroid.
            - easy retains representative items closest to the centroid by sorting points by increasing distance from the centroid.
            - random retains items randomly.
        batched_cosine_similarity (Union[bool, int]): Whether to use batched cosine similarity (has less memory usage).
            Default is 1024. When False or 0, no batching is used and memory requirements are O(N^2) where N is the number of items in the cluster.
            When True, batch size is set to 1024 and memory requirements are O(N*B) where N is the number of items in the cluster and B is the batch size.
        clustering_input_partition_size (Optional[str]): The size of data partition with which to run KMeans.
            Default is "2gb". If None, then the dataset is not repartitioned.

        eps_to_extract (float): Epsilon value to extract deduplicated data.
            Default is 0.01.
    """

    cache_dir: str
    profile_dir: Optional[str] = None
    num_files: int = -1

    # Embeddings
    embedding_model_name_or_path: str = "sentence-transformers/all-MiniLM-L6-v2"
    embedding_batch_size: int = 128
    embeddings_save_loc: str = "embeddings"
    embedding_max_mem_gb: Optional[int] = None

    # Options: "mean_pooling", "last_token"
    embedding_pooling_strategy: str = "mean_pooling"
    embedding_column: str = "embeddings"
    write_embeddings_to_disk: bool = True
    write_to_filename: bool = False

    # Clustering
    max_iter: int = 100
    n_clusters: int = 1000
    clustering_save_loc: str = "clustering_results"
    random_state: int = 1234
    sim_metric: Literal["cosine", "l2"] = "cosine"
    which_to_keep: Literal["hard", "easy", "random"] = "hard"
    batched_cosine_similarity: Union[bool, int] = 1024
    clustering_input_partition_size: str = "2gb"

    # Extract dedup config
    eps_to_extract: float = 0.01

    def __post_init__(self):
        if self.cache_dir is None:
            raise ValueError(
                "Finding sem-dedup requires a cache directory accessible via all workers to store intermediates"
            )
        assert (
            0 <= self.eps_to_extract <= 1
        ), "Epsilon to extract must be between [0, 1]"

        # Convert bool to int
        if isinstance(self.batched_cosine_similarity, bool):
            if self.batched_cosine_similarity:
                self.batched_cosine_similarity = 1024
            else:
                self.batched_cosine_similarity = 0
        if not isinstance(self.batched_cosine_similarity, int):
            raise ValueError("batched_cosine_similarity must be an integer")



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import usaddress
from presidio_analyzer import LocalRecognizer, RecognizerResult


class AddressRecognizer(LocalRecognizer):
    """
    Address Detector based on usaddress library

    Typical US address format:
    * The recipient’s first and last name
    * Street number and name (address line 1)
    * Apartment or unit and its number (address line 2)
    * City, state and zip code (include all of this on one line with a comma between city and state, but not zip code)
    * Country
    """

    def load(self) -> None:
        pass

    def analyze(self, text, entities, nlp_artifacts):
        output = usaddress.parse(text)
        curr_pos = 0
        results = []

        for token, _type in output:
            token_pos_start = text.find(token, curr_pos)
            token_pos_end = token_pos_start + len(token)
            curr_pos = token_pos_end + 1
            if _type != "Recipient" and not _type.startswith("USPS"):
                result = RecognizerResult(
                    entity_type="ADDRESS",
                    start=token_pos_start,
                    end=token_pos_end,
                    score=1,
                )
                results.append(result)
        return results



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import time

from nemo_curator.modules.fuzzy_dedup._shuffle import _Shuffle
from nemo_curator.utils.distributed_utils import get_client, get_num_workers
from nemo_curator.utils.fuzzy_dedup_utils.io_utils import (
    get_text_ddf_from_json_path_with_blocksize,
)
from nemo_curator.utils.script_utils import ArgumentHelper


def func():
    import cudf

    from nemo_curator.modules.fuzzy_dedup._shuffle import _Shuffle


def main(args):
    input_data_paths = args.input_data_dirs
    input_anchor_docs_with_bk_dir = args.input_bucket_mapping_dir
    OUTPUT_PATH = args.output_dir
    output_shuffled_docs_path = os.path.join(OUTPUT_PATH, "shuffled_docs.parquet")

    client = get_client(**ArgumentHelper.parse_client_args(args))
    client.run(func)
    print(f"Num Workers = {get_num_workers(client)}", flush=True)
    print("Connected to dask cluster", flush=True)
    print("Running jaccard shuffle script", flush=True)
    print(f"Args = {args}")
    st = time.time()
    text_ddf = get_text_ddf_from_json_path_with_blocksize(
        input_data_paths=input_data_paths,
        num_files=args.num_files,
        blocksize=args.text_ddf_blocksize,
        id_column=args.input_json_id_field,
        text_column=args.input_json_text_field,
        input_meta=args.input_meta,
    )
    print(
        "Graph creation for get_text_ddf_from_json_path_with_blocksize complete.",
        flush=True,
    )
    print(f"text_ddf.npartitions  = {text_ddf.npartitions}", flush=True)
    shuffle = _Shuffle(
        id_fields=["dataset_id", "doc_id"],
        text_field=args.input_json_text_field,
        profile_dir=args.profile_path,
        int_to_str_id=args.input_json_id_field,
    )
    shuffle.shuffle_docs_on_buckets(
        documents_df=text_ddf,
        bucket_w_anchors_path=input_anchor_docs_with_bk_dir,
        output_shuffled_docs_path=output_shuffled_docs_path,
        bucket_mapping_df_blocksize=args.bucket_mapping_ddf_blocksize,
        parts_per_worker=args.parts_per_worker,
        bucket_parts_per_worker=args.bucket_parts_per_worker,
        partition_on="_output_partition_id",
    )
    et = time.time()
    print(f"Jaccard Shuffle E2E time taken = {et-st} s")


def attach_args():
    description = """
    Shuffles input text documents based on the given bucket
    map. The output is a partitioned Parquet dataset with the documents
    shuffled by buckets.
    """
    parser = argparse.ArgumentParser(
        description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.parse_gpu_dedup_args()

    argumentHelper.add_arg_input_meta()
    argumentHelper.add_arg_output_dir()
    argumentHelper.add_arg_text_ddf_blocksize()
    parser.add_argument(
        "--bucket-mapping-ddf-blocksize",
        type=int,
        default=256,
        help="The block size for for anchor_docs_with_bk DataFrame in MB.",
    )
    parser.add_argument(
        "--bucket-parts-per-worker",
        default=8,
        type=int,
        help="The number of bucket parts to process per worker per batch.",
    )
    parser.add_argument(
        "--input-bucket-mapping-dir",
        type=str,
        help="The directory containing anchor_docs_with_bk files.",
    )
    parser.add_argument(
        "--parts-per-worker",
        default=1,
        type=int,
        help="The number of parts to process per worker per batch.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import time

from nemo_curator.classifiers import MultilingualDomainClassifier
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    global_st = time.time()

    # Input can be a string or list
    input_file_path = "/path/to/data"
    output_file_path = "./"

    client_args = ArgumentHelper.parse_client_args(args)
    client_args["cluster_type"] = "gpu"
    client = get_client(**client_args)

    input_dataset = DocumentDataset.read_json(
        input_file_path, backend="cudf", add_filename=True
    )

    multilingual_domain_classifier = MultilingualDomainClassifier(
        filter_by=["Games", "Sports"]
    )
    result_dataset = multilingual_domain_classifier(dataset=input_dataset)

    result_dataset.to_json(output_path=output_file_path, write_to_filename=True)

    global_et = time.time()
    print(
        f"Total time taken for multilingual domain classifier inference: {global_et-global_st} s",
        flush=True,
    )

    client.close()


def attach_args(
    parser=argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ),
):
    argumentHelper = ArgumentHelper(parser)
    argumentHelper.add_distributed_classifier_cluster_args()

    return argumentHelper.parser


if __name__ == "__main__":
    main(attach_args().parse_args())



import json
import logging
import os

import dask.dataframe as dd
from dask.distributed import Client, LocalCluster

from nemo_curator.utils.distributed_utils import get_num_workers

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)

DATA_BASE = os.environ.get("DATA_BASE")
CC_BASE = os.path.join(DATA_BASE, "fuzzy/cc/")
CC_FOLDER = os.path.join(CC_BASE, "connected_components.parquet")
CC_CONVERTED_FOLDER = os.path.join(CC_BASE, "connected_components_converted.parquet")
ID_MAPPING = os.path.join(DATA_BASE, "dataset_id_mapping.json")
CPU_WORKERS = os.environ.get("CPU_WORKERS")


if __name__ == "__main__":
    cluster = LocalCluster(n_workers=CPU_WORKERS, processes=True)
    client = Client(cluster)
    logging.info(f"Number of dask workers: {get_num_workers(client)}")

    cc_df = dd.read_parquet(CC_FOLDER, split_row_groups=False)

    with open(ID_MAPPING, "r") as f:
        dataset_id_mapping = json.loads(f.read())

    global_dataset_id_mapping = {}
    for key, val in dataset_id_mapping.items():
        if "dclm" in val:
            global_dataset_id_mapping[key] = "dclm"
        elif "fwe2" in val:
            global_dataset_id_mapping[key] = "fwe2"
        elif "zyda" in val:
            global_dataset_id_mapping[key] = "zyda"
        elif "dolma-cc" in val:
            global_dataset_id_mapping[key] = "dolma-cc"
        else:
            print(f"Unknown value {val} for key {key}")

    def convert_cc_ids(
        cc_df, dataset_id_mapping, global_dataset_id_mapping, doc_id_len=10
    ):
        cc_df["global_dataset_id"] = cc_df.dataset_id.astype(str).replace(
            global_dataset_id_mapping
        )
        cc_df["dataset_id"] = cc_df.dataset_id.astype(str).replace(dataset_id_mapping)
        cc_df["doc_id"] = (
            cc_df["doc_id"]
            .astype(str)
            .str.pad(width=doc_id_len, side="left", fillchar="0")
        )
        cc_df["original_id"] = cc_df.dataset_id + "-" + cc_df.doc_id
        return cc_df[["global_dataset_id", "dataset_id", "original_id", "group"]]

    cc_df_converted = convert_cc_ids(
        cc_df, dataset_id_mapping, global_dataset_id_mapping
    )
    cc_df_converted.to_parquet(CC_CONVERTED_FOLDER, overwrite=True, write_index=False)
    logging.info("Done converting!")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from nemo_curator import (
    ExactDuplicates,
    FuzzyDuplicates,
    FuzzyDuplicatesConfig,
    Modify,
    ScoreFilter,
    SemDedup,
    SemDedupConfig,
    Sequential,
)
from nemo_curator.datasets import DocumentDataset
from nemo_curator.filters import (
    DocumentFilter,
    RepeatedParagraphsFilter,
    RepeatingTopNGramsFilter,
    UrlsFilter,
    WordCountFilter,
)
from nemo_curator.filters.code import (
    NumberOfLinesOfCodeFilter,
    PythonCommentToCodeFilter,
)
from nemo_curator.modifiers import DocumentModifier
from nemo_curator.modifiers.pii_modifier import PiiModifier
from nemo_curator.modifiers.unicode_reformatter import UnicodeReformatter
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir


class QuotationUnifier(DocumentModifier):
    """
    A simple modifier that unifies the quotation marks in the documents.
    """

    def modify_document(self, text: str) -> str:
        """
        Modifies the given text by replacing left and right single quotes with normal single quotes,
        and replacing left and right double quotes with normal double quotes.

        Args:
            text (str): The text to be modified.

        Returns:
            str: The modified text.
        """
        text = text.replace("‘", "'").replace("’", "'")
        text = text.replace("“", '"').replace("”", '"')
        return text


def clean_and_unify(dataset: DocumentDataset) -> DocumentDataset:
    """
    Cleans and unifies the given dataset using a set of predefined cleaners.

    Args:
        dataset (DocumentDataset): The dataset to be cleaned and unified.

    Returns:
        DocumentDataset: The cleaned and unified dataset.
    """
    cleaners = Sequential(
        [
            # Unify all the quotation marks
            Modify(QuotationUnifier(), text_field="text"),
            # Unify all unicode
            Modify(UnicodeReformatter(), text_field="text"),
        ]
    )
    return cleaners(dataset)


def filter_text(dataset: DocumentDataset) -> DocumentDataset:
    """
    Filters the given dataset based on various criteria.
    Refer to the full list of all filters here:
    https://github.com/NVIDIA/NeMo-Curator/blob/main/config/heuristic_filter_en.yaml
    https://github.com/NVIDIA/NeMo-Curator/blob/main/tutorials/peft-curation/main.py

    Args:
        dataset (DocumentDataset): The dataset to be filtered.

    Returns:
        DocumentDataset: The filtered dataset.
    """
    filters = Sequential(
        [
            # If a document contains a number of words not
            # within a specified range then discard
            ScoreFilter(
                WordCountFilter(min_words=50, max_words=100000),
                text_field="text",
                score_field="word_count",
                score_type=int,
            ),
            # If the document shrinks by > x% in terms of number of characters after
            # removing the top n-grams then discard. Source: Gopher (Rae et al., 2021)
            ScoreFilter(
                RepeatingTopNGramsFilter(n=2, max_repeating_ngram_ratio=0.2),
                text_field="text",
                score_type=float,
            ),
            ScoreFilter(
                RepeatingTopNGramsFilter(n=3, max_repeating_ngram_ratio=0.18),
                text_field="text",
                score_type=float,
            ),
            ScoreFilter(
                RepeatingTopNGramsFilter(n=4, max_repeating_ngram_ratio=0.16),
                text_field="text",
                score_type=float,
            ),
            ScoreFilter(
                RepeatedParagraphsFilter(max_repeated_paragraphs_ratio=0.7),
                text_field="text",
                score_type=float,
            ),
            # If more than 20% of the document is comprised of URLs then discard
            ScoreFilter(
                UrlsFilter(max_url_to_text_ratio=0.2),
                text_field="text",
                score_type=float,
            ),
        ]
    )
    filtered_dataset = filters(dataset)
    return filtered_dataset


def filter_code(dataset: DocumentDataset) -> DocumentDataset:
    """
    Filters the given dataset based on various criteria.
    Refer to the full list of all filters here:
    https://github.com/NVIDIA/NeMo-Curator/blob/main/config/heuristic_filter_en.yaml
    https://github.com/NVIDIA/NeMo-Curator/blob/main/tutorials/peft-curation/main.py

    Args:
        dataset (DocumentDataset): The dataset to be filtered.

    Returns:
        DocumentDataset: The filtered dataset.
    """
    filters = Sequential(
        [
            # If a document contains a number of words not
            # within a specified range then discard
            ScoreFilter(
                WordCountFilter(min_words=5, max_words=100000),
                text_field="text",
                score_field="word_count",
                score_type=int,
            ),
        ]
    )
    filtered_dataset = filters(dataset)
    return filtered_dataset


def filter_code_dataset(dataset: DocumentDataset) -> DocumentDataset:
    """
    Filters the given code dataset based on various code-specific criteria.
    Refer to the full list of all filters here:
    https://github.com/NVIDIA/NeMo-Curator/blob/main/config/heuristic_filter_code.yaml
    https://github.com/NVIDIA/NeMo-Curator/blob/main/nemo_curator/filters/code.py

    Args:
        dataset (DocumentDataset): The dataset to be filtered.

    Returns:
        DocumentDataset: The filtered dataset.
    """
    filters = Sequential(
        [
            # if the comment to code ratio is not within the specified range,
            # discard the document
            ScoreFilter(
                PythonCommentToCodeFilter(
                    min_comment_to_code_ratio=0.001, max_comment_to_code_ratio=0.80
                ),
                text_field="text",
                score_type=float,
            ),
            # if the number of lines of code is not within the specified range,
            # discard the document
            ScoreFilter(
                NumberOfLinesOfCodeFilter(min_lines=5, max_lines=20000),
                text_field="text",
                score_type=int,
            ),
        ]
    )
    filtered_dataset = filters(dataset)
    return filtered_dataset


def redact_pii(dataset: DocumentDataset) -> DocumentDataset:
    """
    Redacts personally identifiable information (PII) from extracted comment lines in a given dataset.

    Args:
        dataset (DocumentDataset): The dataset containing documents with PII.

    Returns:
        DocumentDataset: The redacted dataset with PII replaced by a generic value.
    """
    DEFAULT_MAX_DOC_SIZE = 3000000
    redactor = Modify(
        PiiModifier(
            supported_entities=[
                "PERSON",
                "EMAIL_ADDRESS",
            ],
            anonymize_action="replace",
            device="gpu",
        ),
        text_field="extracted_comment",
    )
    return redactor(dataset)


def redact_code(dataset: DocumentDataset) -> DocumentDataset:
    """
    Redact PII on comment lines on a  given code dataset.

    Args:
        dataset (DocumentDataset): The dataset to be redacted.

    Returns:
        DocumentDataset: The redacted dataset.
    """

    # functions to extract comment lines from each row in a dataframe
    def func(row):
        return row["text"][row["text"].find("/*") : row["text"].find("*/") + 2]

    def func2(row):
        comment = row["text"][row["text"].find("/*") : row["text"].find("*/") + 2]
        return row["text"].replace(comment, str(row["extracted_comment"]))

    dataset.df["extracted_comment"] = dataset.df.apply(func, axis=1, meta=(None, str))
    redacted_dataset = redact_pii(dataset)
    redacted_dataset.df["text"] = redacted_dataset.df.apply(
        func2, axis=1, meta=(None, str)
    )
    redacted_dataset.df = redacted_dataset.df.drop(["extracted_comment"], axis=1)

    return redacted_dataset


def exact_dedupe(dataset: DocumentDataset) -> DocumentDataset:
    """
    Remove exact duplicates from the given DocumentDataset.

    Args:
        dataset (DocumentDataset): The dataset containing documents.

    Returns:
        DocumentDataset: The deduplicated dataset.
    """
    deduplicator = ExactDuplicates(id_field="id", text_field="text", hash_method="md5")
    # Find the duplicates
    duplicates = deduplicator(dataset)
    docs_to_remove = duplicates.df.map_partitions(
        lambda x: x[x._hashes.duplicated(keep="first")]
    )
    # Remove the duplicates using their IDs.
    duplicate_ids = list(docs_to_remove.compute().id)
    dataset_df = dataset.df
    deduped = dataset_df[~dataset_df.id.isin(duplicate_ids)]
    return DocumentDataset(deduped)


def fuzzy_dedupe(dataset: DocumentDataset, cache_dir: str) -> DocumentDataset:
    """
    Removes near-duplicate documents and code lines

    Args:
        dataset (DocumentDataset): The dataset containing documents.
        cache_dir (str): Directory for storing intermediate results.

    Returns:
        DocumentDataset: The deduplicated dataset.
    """
    fuzzy_dedup_config = FuzzyDuplicatesConfig(
        cache_dir=cache_dir,
        id_field="id",
        text_field="text",
        seed=42,
        char_ngrams=24,
        num_buckets=20,
        hashes_per_bucket=13,
        use_64_bit_hash=False,
        buckets_per_shuffle=5,
        false_positive_check=False,
    )
    fuzzy_dup = FuzzyDuplicates(config=fuzzy_dedup_config)
    duplicates = fuzzy_dup(dataset)
    if duplicates is None:
        return dataset
    else:
        docs_to_remove = duplicates.df.map_partitions(
            lambda x: x[x.group.duplicated(keep="first")]
        )

        # When there are few duplicates we can compute the results to a list and use `isin`.
        duplicate_ids = docs_to_remove.compute().id.to_arrow().to_pylist()
        dataset_df = dataset.df
        deduped = dataset_df[~dataset_df.id.isin(duplicate_ids)]
        return DocumentDataset(deduped)


def semantic_dedupe(
    dataset: DocumentDataset,
    sem_dedupe_config_yaml_path: str,
):
    """
    Perform semantic deduplication on the given dataset.

    Args:
        dataset (DocumentDataset): The dataset containing documents.
        sem_dedupe_config_yaml_path (str): The path to the semantic dedupe configuration file.

    Returns:
        The deduplicated DocumentDataset.
    """
    partition_lengths = dataset.df.map_partitions(len).compute()
    non_empty_partitions = [
        i for i, length in enumerate(partition_lengths) if length > 0
    ]
    dataset.df = dataset.df.partitions[non_empty_partitions]

    semdedup_config = SemDedupConfig.from_yaml(sem_dedupe_config_yaml_path)
    expand_outdir_and_mkdir(semdedup_config.cache_dir)
    semdup = SemDedup(config=semdedup_config, perform_removal=True)
    deduplicated_dataset = semdup(dataset)
    return deduplicated_dataset


class TextLineCountFilter(DocumentFilter):
    """
    Discard text files based on number of lines.
    """

    def __init__(self, min_lines: int = 10):
        super().__init__()
        self._min_lines = min_lines

    def score_document(self, text: str) -> bool:
        words = text.split()
        if words[0] == "text" and int(words[2]) < self._min_lines:
            return False
        else:
            return True

    def keep_document(self, score) -> bool:
        return score


class CodeLineCountFilter(DocumentFilter):
    """
    Discard code files based on number of lines.
    """

    def __init__(self, min_lines: int = 10, max_lines: int = 20000):
        super().__init__()
        self._min_lines = min_lines
        self._max_lines = max_lines

    def score_document(self, text: str) -> bool:
        words = text.split()
        if words[0] == "code" and (
            int(words[2]) < self._min_lines or int(words[2]) > self._max_lines
        ):
            return False
        else:
            return True

    def keep_document(self, score) -> bool:
        return score


def rm_dir(cache_dir: str):
    if os.path.isdir(cache_dir):
        os.system(f"rm -rf {cache_dir}")



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from abc import ABC, abstractmethod

from nemo_curator.utils.text_utils import get_words


class DownstreamTask(ABC):

    def __init__(self):
        super().__init__()
        self._task_name = None
        self._ngrams = {}

    @abstractmethod
    def generate_ngrams(self):
        pass

    @property
    def ngrams(self):
        return self._ngrams

    def _update_ngrams(self, text, min_ngram_size=8, max_ngram_size=13):
        words, positions = get_words(text)
        if len(words) < min_ngram_size:
            return

        if len(words) < max_ngram_size:
            seq = " ".join(words)
            if seq not in self._ngrams:
                self._ngrams[seq] = 0

        for i in range(len(words) - max_ngram_size + 1):
            seq = " ".join(words[i : i + max_ngram_size])
            if seq not in self._ngrams:
                self._ngrams[seq] = 0


def import_task(task_path):
    module_path, task_name = task_path.rsplit(".", 1)
    task_module = importlib.import_module(module_path)
    task_class = getattr(task_module, task_name)
    if not issubclass(task_class, DownstreamTask):
        raise ValueError(
            f"Input iterator {task_class.__name__} "
            "must be derived from DownstreamTask"
            "defined in nemo_curator.tasks.downstream_task"
        )
    return task_class



import os

from nemo_curator import AddId
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.file_utils import get_all_files_paths_under


def ensure_directory_exists(filename: str):
    os.makedirs(os.path.dirname(filename), exist_ok=True)


def process_data(input_folder, output_folder, prefix, partition_size="512MB"):
    raw_files = get_all_files_paths_under(input_folder)
    raw_data = DocumentDataset.read_parquet(raw_files)
    raw_data_rep = DocumentDataset(
        raw_data.df.repartition(partition_size=partition_size)
    )
    add_id = AddId(id_field="nemo_id", id_prefix=prefix)
    data_with_id = add_id(raw_data_rep)
    ensure_directory_exists(output_folder)
    data_with_id.to_parquet(output_folder)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import random

import nemo_curator
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client, read_data, write_to_disk
from nemo_curator.utils.file_utils import (
    expand_outdir_and_mkdir,
    get_all_files_paths_under,
)
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))

    backend = "cudf" if args.device == "gpu" else "pandas"
    output_dir = expand_outdir_and_mkdir(args.output_data_dir)
    files = get_all_files_paths_under(args.input_data_dir)
    if args.shuffle:
        random.seed(args.seed)
        random.shuffle(files)

    dataset = DocumentDataset(
        read_data(
            files, file_type=args.input_file_type, backend=backend, add_filename=True
        )
    )
    add_id = nemo_curator.AddId(
        args.id_field_name, id_prefix=args.id_prefix, start_index=args.starting_index
    )
    id_dataset = add_id(dataset)

    write_to_disk(
        id_dataset.df,
        output_dir,
        write_to_filename=True,
        output_type=args.output_file_type,
    )


def attach_args(
    parser=argparse.ArgumentParser(
        """
Adds unique identifiers to each document in the dataset.
Creates a new ID field with name specified by the argument
"--id-field-name" within each JSON file.

This script essentially works by counting the total
number of documents within the dataset, and then in parallel
assigns unique sequential IDs to each document in the dataset.

If a document identifier does not already exist for each document, then
these IDs must be added prior to performing fuzzy and/or exact deduplication.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_input_data_dir()
    argumentHelper.add_arg_input_file_type()
    argumentHelper.add_arg_output_data_dir(
        help="The output directory to where the JSONL files with IDs will "
        "be written."
    )
    argumentHelper.add_arg_output_file_type()
    argumentHelper.add_arg_seed()
    argumentHelper.add_arg_shuffle(
        help="Shuffle the order of files before assigning IDs. "
        "Useful for creating a copy dataset with different IDs."
    )
    argumentHelper.add_distributed_args()
    argumentHelper.set_default_n_workers(2.5)
    parser.add_argument(
        "--id-field-name",
        type=str,
        required=True,
        help="The name of the field that will contain the id value. "
        "This is a required argument.",
    )
    parser.add_argument(
        "--id-prefix",
        type=str,
        default="doc_id",
        help="The prefix to the ID number that will be assigned to the "
        "document. When performing deduplication jointly with different"
        "datasets, it is helpful to provide a prefix that denotes that a "
        "document belongs to a particular dataset (e.g., wiki for documents"
        "that come from the Wikipedia dataset).",
    )
    parser.add_argument(
        "--starting-index",
        type=int,
        default=None,
        help="If supplied, determines the starting index from which to start "
        "indexing the documents. By default, it is unspecified, and uses an ID"
        " scheme that is fast to calculate and is not guaranteed to be ordered.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



import math
from typing import Callable, List, Optional

import dask.dataframe as dd
import numpy as np

from nemo_curator.datasets.doc_dataset import DocumentDataset
from nemo_curator.modules.base import BaseModule


def default_filename(partition_num: int) -> str:
    return f"file_{partition_num:010d}.jsonl"


class Shuffle(BaseModule):
    def __init__(
        self,
        seed: Optional[int] = None,
        npartitions: Optional[int] = None,
        partition_to_filename: Callable[[int], str] = default_filename,
        filename_col: str = "file_name",
    ) -> None:
        """
        Randomly permutes the dataset. This will make the original filename_col column invalid, so if the column is present it will be overwritten.
        Args:
            seed: The random seed that will be used to determine which partition (file) each datapoint goes to.
                Setting the seed will guarantee determinism, but may be slightly slower (20-30% slower)
                depending on the dataset size.
            npartitions: The output number of partitions to create in the dataset.
                If None, it will retain the same number of partitions as the original dataset.
            partition_to_filename: If the filename column is present, it will be overwritten.
                Passing a function in through this argument allows the user to configure what the filename
                will look like given the partition number. The default method names the partition
                f'file_{partition_num:010d}.jsonl' and should be changed if the user is not using a .jsonl format.
        """
        super().__init__(input_backend="pandas")
        self.seed = seed
        self.npartitions = npartitions
        self.partition_to_filename = partition_to_filename
        self.rand_col = "_shuffle_rand"
        self.filename_col = filename_col

    def call(self, dataset: DocumentDataset) -> DocumentDataset:
        if self.seed is None:
            return self.shuffle_nondeterministic(dataset)
        else:
            return self.shuffle_deterministic(dataset)

    def shuffle_deterministic(self, dataset: DocumentDataset) -> DocumentDataset:
        new_npartitions = (
            dataset.df.npartitions if self.npartitions is None else self.npartitions
        )

        dataset.df[self.rand_col] = dataset.df.map_partitions(self._add_rand_col)

        shuffled_df = dataset.df.set_index(self.rand_col, npartitions=new_npartitions)
        shuffled_df = shuffled_df.reset_index(drop=True)

        if self.filename_col in shuffled_df:
            shuffled_df[self.filename_col] = shuffled_df.map_partitions(
                self._add_filename
            )

        return DocumentDataset(shuffled_df)

    def shuffle_nondeterministic(self, dataset: DocumentDataset) -> DocumentDataset:
        new_npartitions = (
            dataset.df.npartitions if self.npartitions is None else self.npartitions
        )

        dataset.df[self.rand_col] = dataset.df.map_partitions(self._add_rand_col)

        shuffled_df = dataset.df.shuffle(
            self.rand_col, npartitions=new_npartitions, ignore_index=True
        )
        shuffled_df = shuffled_df.drop(columns=[self.rand_col])
        shuffled_df = shuffled_df.map_partitions(self._partition_shuffle)

        return DocumentDataset(shuffled_df)

    def _add_rand_col(self, partition, partition_info=None):
        if partition_info is None:
            partition_info = {
                "number": 0,
            }

        if self.seed is not None:
            np.random.seed(self.seed + partition_info["number"])
        rand_col = np.random.randint(0, np.iinfo("int64").max, size=len(partition))

        return rand_col

    def _partition_shuffle(self, partition, partition_info=None):
        if partition_info is None:
            return partition

        partition_num = partition_info["number"]
        if self.seed is not None:
            random_state = self.seed + partition_num
        else:
            random_state = None

        partition = partition.sample(frac=1, random_state=random_state).reset_index(
            drop=True
        )

        if self.filename_col in partition:
            filename = self.partition_to_filename(partition_num)
            partition[self.filename_col] = filename

        return partition

    def _add_filename(self, partition, partition_info=None):
        if partition_info is None:
            return [self.filename_col] * len(partition)

        filename = self.partition_to_filename(partition_info["number"])

        return [filename for _ in range(len(partition))]


def blend_datasets(
    target_size: int, datasets: List[DocumentDataset], sampling_weights: List[float]
) -> DocumentDataset:
    """
    Combines multiple datasets into one with different amounts of each dataset.
    Args:
        target_size: The number of documents the resulting dataset should have.
            The actual size of the dataset may be slightly larger if the normalized weights do not allow
            for even mixtures of the datasets.
        datasets: A list of all datasets to combine together
        sampling_weights: A list of weights to assign to each dataset in the input. Weights will be
            normalized across the whole list as a part of the sampling process. For example, if the normalized
            sampling weight for dataset 1 is 0.02, 2% ofthe total samples will be sampled from dataset 1.
            There are guaranteed to be math.ceil(normalized_weight_i * target_size) elements from dataset i in
            the final blend.
    """
    if len(datasets) != len(sampling_weights):
        raise ValueError(
            f"Different number of datasets and weights specified. {len(datasets)} datasets and {len(sampling_weights)}"
        )

    weight_sum = sum(sampling_weights)
    sampling_weights = [weight / weight_sum for weight in sampling_weights]
    num_documents_per_dataset = [
        math.ceil(weight * target_size) for weight in sampling_weights
    ]

    blend_components = []
    for dataset, num_documents in zip(datasets, num_documents_per_dataset):
        # Repeatedly sample from the dataset
        while num_documents > 0:
            sample = _partition_head(dataset.df, num_documents)
            blend_components.append(sample)
            num_documents -= len(sample)

    blended_dataset = dd.concat(blend_components)

    return DocumentDataset(blended_dataset)


def _partition_head(ddf: dd.DataFrame, n: int) -> dd.DataFrame:
    """
    Returns the first n rows in a dataframe while preserving the partitions.
    Meant as a replacement for ddf.head(npartitions=-1, compute=False) as it
    uses too much memory at large scales

    Args:
        ddf: The dataframe to get the first rows from
        n: The number of rows to get
    """
    original_meta = ddf.dtypes.to_dict()
    partition_lengths = ddf.map_partitions(len)
    num_partitions = 0
    total_size = 0
    last_length = 0
    for length in partition_lengths:
        total_size += length
        num_partitions += 1
        last_length = length
        if total_size >= n:
            break

    delayed_df = ddf.to_delayed()
    excess_elems = max(0, total_size - n)
    delayed_df = delayed_df[:num_partitions]
    delayed_df[-1] = delayed_df[-1].head(last_length - excess_elems)

    return dd.from_delayed(delayed_df, meta=original_meta)



import logging
import os
import time

import dask_cudf

from nemo_curator import MinHash
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client, get_num_workers
from nemo_curator.utils.file_utils import get_all_files_paths_under

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)


def read_folder(input_folder, columns=["nemo_id", "text"]):
    data_paths = get_all_files_paths_under(input_folder, keep_extensions="parquet")
    data_paths.sort()
    logging.info(f"Number of files being read: {len(data_paths)}")
    text_ddf = dask_cudf.read_parquet(
        data_paths,
        columns=columns,
    )
    return text_ddf


DATA_BASE = os.environ.get("DATA_BASE")
SCHEDULER_FILE = os.environ.get("SCHEDULER_FILE")


if __name__ == "__main__":
    client = get_client(scheduler_file=SCHEDULER_FILE)
    logging.info(f"Number of dask workers: {get_num_workers(client)}")

    minhash_base_output_path = os.path.join(DATA_BASE, "fuzzy/minhash")
    minhash_output_dir = os.path.join(minhash_base_output_path, "data")

    # Relevant parameters
    minhash_id_field = "nemo_id"
    minhash_text_field = "text"
    seed = 10
    minhash_length = 128
    char_ngram = 25
    use_64bit_hash = False

    # Reading all the data
    text_ddf = read_folder(
        input_folder=os.path.join(DATA_BASE, "processed"),
        columns=[minhash_id_field, minhash_text_field],
    )

    # Computing minhashes
    t0 = time.time()
    minhasher = MinHash(
        seed=seed,
        num_hashes=minhash_length,
        char_ngrams=char_ngram,
        use_64bit_hash=use_64bit_hash,
        id_field=minhash_id_field,
        text_field=minhash_text_field,
        cache_dir=minhash_output_dir,
    )
    res = minhasher(DocumentDataset(text_ddf)).df
    logging.info(f"Time taken for MinHash: {time.time()-t0:.2f}sec.")



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client, read_data, write_to_disk
from nemo_curator.utils.file_utils import (
    expand_outdir_and_mkdir,
    get_all_files_paths_under,
)
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))

    out_dir = expand_outdir_and_mkdir(args.output_data_dir)

    input_dirs = args.input_data_dirs.split(",")
    weights = [float(weight) for weight in args.weights.split(",")]

    datasets = [
        DocumentDataset(
            read_data(
                get_all_files_paths_under(path),
                file_type=args.input_file_type,
                backend="pandas",
            )
        )
        for path in input_dirs
    ]

    output_dataset = nc.blend_datasets(args.target_samples, datasets, weights)

    if args.shuffle:
        shuffle = nc.Shuffle(seed=args.seed)
        output_dataset = shuffle(output_dataset)

    write_to_disk(output_dataset.df, out_dir, output_type=args.output_file_type)

    client.close()


def attach_args(
    parser=argparse.ArgumentParser(
        """
Blends a collection of datasets together based on certain weights.

It takes as input a comma-separated list of dataset directories, the
corresponding weights that should be associated with each datatset,
and the target number of samples to aggregate from across all the datasets.
The file shards of the resulting dataset are not guaranteed to be even
or reflect the original dataset(s).

A blend is created from these datasets and saved to the specified output directory.
Optionally, the user can choose to shuffle this dataset as well.
  """,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_input_file_type()
    argumentHelper.add_arg_output_data_dir(
        help="The output directory to where the blended dataset will be written."
    )
    argumentHelper.add_arg_output_file_type()
    argumentHelper.add_arg_seed()
    argumentHelper.add_arg_shuffle(help="Shuffles the dataset after blending.")
    argumentHelper.add_distributed_args()
    parser.add_argument(
        "--input-data-dirs",
        type=str,
        default=None,
        help="Comma-separated list of directories consisting of dataset "
        "files that are accessible to all nodes.",
    )
    parser.add_argument(
        "--target-samples",
        type=int,
        default=10000,
        help="The number of samples to be included in the output dataset."
        " There may be more samples in order to accurately reflect the "
        "weight balance, but there will never be less.",
    )
    parser.add_argument(
        "--weights",
        type=str,
        default=None,
        help="Comma-separated list of floating-point weights corresponding "
        "to each dataset passed in --input-data-dirs.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



import dask.dataframe as dd
import pandas as pd

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset


def list_to_dataset(documents, col_name="text", npartitions=2):
    data = {col_name: documents}
    pdf = pd.DataFrame(data)

    return DocumentDataset(dd.from_pandas(pdf, npartitions=npartitions))


def all_equal(left_dataset, right_dataset):
    left_result = left_dataset.df.compute()
    right_result = right_dataset.df.compute()

    l_cols = set(left_result.columns)
    r_cols = set(right_result.columns)
    assert l_cols == r_cols
    for col in left_result.columns:
        left = left_result[col].reset_index(drop=True)
        right = right_result[col].reset_index(drop=True)
        assert all(left == right), f"Mismatch in {col} column.\n{left}\n{right}\n"


class TestBlending:
    def test_blend_as_original(self):
        first_dataset = list_to_dataset(["one", "two", "three"])
        result_dataset = nc.blend_datasets(len(first_dataset), [first_dataset], [1.0])
        all_equal(first_dataset, result_dataset)

    def test_equal_blend(self):
        first_dataset = list_to_dataset(["a", "a"])
        second_dataset = list_to_dataset(["b", "b"])
        result_dataset = nc.blend_datasets(
            2, [first_dataset, second_dataset], [0.5, 0.5]
        )
        counts = result_dataset.df["text"].value_counts().compute()
        assert len(result_dataset) == 2
        assert counts["a"] == 1
        assert counts["b"] == 1

    def test_equal_blend_with_weights(self):
        first_dataset = list_to_dataset(["a", "a"])
        second_dataset = list_to_dataset(["b", "b"])
        result_dataset = nc.blend_datasets(
            2, [first_dataset, second_dataset], [2.0, 2.0]
        )
        counts = result_dataset.df["text"].value_counts().compute()
        assert len(result_dataset) == 2
        assert counts["a"] == 1
        assert counts["b"] == 1

    def test_uneven_blend(self):
        first_dataset = list_to_dataset(["a", "a"])
        second_dataset = list_to_dataset(["b", "b"])
        result_dataset = nc.blend_datasets(
            4, [first_dataset, second_dataset], [3.0, 1.0]
        )
        counts = result_dataset.df["text"].value_counts().compute()
        assert len(result_dataset) == 4
        assert counts["a"] == 3
        assert counts["b"] == 1

    def test_very_uneven_blend(self):
        first_dataset = list_to_dataset(["a", "a"])
        second_dataset = list_to_dataset(["b", "b"])
        result_dataset = nc.blend_datasets(
            4, [first_dataset, second_dataset], [1.0, 0.0]
        )
        counts = result_dataset.df["text"].value_counts().compute()
        assert len(result_dataset) == 4
        assert counts["a"] == 4
        assert "b" not in counts

    def test_proper_uneven_blend(self):
        first_dataset = list_to_dataset(["a", "b", "c", "d"])
        second_dataset = list_to_dataset(["e", "f"])
        result_dataset = nc.blend_datasets(
            8, [first_dataset, second_dataset], [1.0, 0.0]
        )
        counts = result_dataset.df["text"].value_counts().compute()
        assert len(result_dataset) == 8
        assert counts["a"] == 2
        assert counts["b"] == 2
        assert counts["c"] == 2
        assert counts["d"] == 2

    def test_four_dataset_blend(self):
        datasets = []
        datasets.append(list_to_dataset(["a", "a"]))
        datasets.append(list_to_dataset(["b", "b", "b"]))
        datasets.append(list_to_dataset(["c"]))
        datasets.append(list_to_dataset(["d", "d", "d", "d"]))
        result_dataset = nc.blend_datasets(8, datasets, [1.0, 2.0, 3.0, 4.0])
        counts = result_dataset.df["text"].value_counts().compute()
        assert len(result_dataset) == 10
        assert counts["a"] == 1
        assert counts["b"] == 2
        assert counts["c"] == 3
        assert counts["d"] == 4



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import shutil
from typing import Any, Optional, Tuple

import matplotlib.pyplot as plt
from downloaders import (
    download_github_sources,
    download_pdf_sources,
    download_wikipedia_sources,
)
from utils import (
    CodeLineCountFilter,
    TextLineCountFilter,
    clean_and_unify,
    exact_dedupe,
    filter_code,
    filter_text,
    fuzzy_dedupe,
    redact_code,
    rm_dir,
    semantic_dedupe,
)

import nemo_curator as nc
from nemo_curator import ScoreFilter, Sequential
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.file_utils import (
    get_all_files_paths_under,
    separate_by_metadata,
)
from nemo_curator.utils.script_utils import ArgumentHelper

SCRIPT_DIR_PATH = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(SCRIPT_DIR_PATH, "data")
CONFIG_DIR = os.path.join(SCRIPT_DIR_PATH, "configs")


def download_sources(
    wikipedia_limit: Optional[int] = None,
    github_limit: Optional[int] = None,
    pdf_limit: Optional[int] = None,
) -> Tuple[str, str]:
    """
    Downloads all the dataset sources and converts them to the JSONL format.
    Args:
        wikipedia_limit (int): Maximum number of wiki urls to be downloaded
        github_limit (int): Maximum number of github repos to be downloaded
        pdf_limit (int): Maximum number of pdf to be downloaded


    Returns:
        tuple: the list of text files and the list of code files.
    """

    wikipedia_dir = download_wikipedia_sources(
        "sources/wikipedia_urls.jsonl", limit=wikipedia_limit
    )
    github_dir = download_github_sources(
        "sources/github_repos.jsonl", limit=github_limit
    )
    pdf_dir = download_pdf_sources("sources/arxiv_urls.jsonl", limit=pdf_limit)

    wiki_files = get_all_files_paths_under(wikipedia_dir)
    code_files = get_all_files_paths_under(github_dir)
    pdf_files = get_all_files_paths_under(pdf_dir)

    text_files = wiki_files + pdf_files

    return text_files, code_files


def plot_data(orig_dataset: DocumentDataset, filename: str):
    """
    Plot histogram of different file types and corresponding sizes

    Args:
        dataset (DocumentDataset): Dataset
        filename (str): Name of the plot to be saved ('sample.png')
    Returns:
        None (saves the plotted file in current directory)
    """
    # visualize file types and sizes
    orig_df = orig_dataset.df.compute()
    orig_df = orig_df.reset_index()

    # Create a histogram for different file types -text
    fig, ax = plt.subplots(figsize=(10, 6))
    orig_df.groupby("file_extension")["size_in_bytes"].sum().plot(kind="bar", ax=ax)
    ax.set_xlabel("file_extension")
    ax.set_ylabel("size_in_bytes")
    ax.set_title("File Size Histogram by File Extension")

    # Save the histogram to a file
    fig.savefig(filename, bbox_inches="tight")


def run_curation_pipeline(args: Any, text_files: str, code_files: str) -> None:
    """
    Run the curation pipeline on the Wiki+Arxiv+Github datasets.

    Args:
        args (Any): Command-line arguments.
        jsonl_dir (str): Directory path where the JSONL files are stored.
    """
    # Initialize the Dask cluster.
    client = get_client(
        **ArgumentHelper.parse_client_args(args), set_torch_to_use_rmm=True
    )

    # Define data curation steps for text and pdf files
    curation_steps_text = Sequential(
        [
            clean_and_unify,
            ScoreFilter(
                TextLineCountFilter(), text_field="file_type_count", score_type=bool
            ),
            filter_text,
            exact_dedupe,
        ]
    )

    # Define data curation steps for code files
    curation_steps_code = Sequential(
        [
            clean_and_unify,
            ScoreFilter(
                CodeLineCountFilter(), text_field="file_type_count", score_type=bool
            ),
            filter_code,
            exact_dedupe,
            redact_code,
        ]
    )

    orig_dataset_text = DocumentDataset.read_json(text_files, add_filename=True)
    orig_dataset_code = DocumentDataset.read_json(code_files, add_filename=True)

    # Create a histogram for different file types -text
    plot_data(orig_dataset_text, "file_size_histogram_txt.png")

    # Create a histogram for different file types - code
    plot_data(orig_dataset_code, "file_size_histogram_code.png")

    # create a field combining fields file type and line count
    orig_dataset_text.df["file_type_count"] = (
        orig_dataset_text.df["file_type"]
        + " : "
        + orig_dataset_text.df["line_count"].astype(str)
    )
    orig_dataset_code.df["file_type_count"] = (
        orig_dataset_code.df["file_type"]
        + " : "
        + orig_dataset_code.df["line_count"].astype(str)
    )

    print("Executing the curation pipeline...")
    dataset_text = curation_steps_text(orig_dataset_text)
    dataset_code = curation_steps_code(orig_dataset_code)

    print("********************* Generating Statistics *********************")
    print(f"Original dataset length for text files: {len(orig_dataset_text.df)}")
    print(f"After dataprep for text files: {len(dataset_text.df)}")
    print(f"Original dataset length for code files: {len(orig_dataset_code.df)}")
    print(f"After dataprep length for code files: {len(dataset_code.df)}")

    if args.device == "gpu":
        print("Executing the semantic dedupe pipeline...")
        gpu_dataset_text = DocumentDataset(dataset_text.df.to_backend("cudf"))
        gpu_dataset_code = DocumentDataset(dataset_code.df.to_backend("cudf"))
        sem_dedupe_config_yaml_path = os.path.join(
            CONFIG_DIR, "text_semantic_dedupe_config.yaml"
        )
        CACHE_DIR = os.path.join(SCRIPT_DIR_PATH, "cache", "semantic_dedupe", "text")
        rm_dir(CACHE_DIR)
        semantic_dataset_text = semantic_dedupe(
            dataset=gpu_dataset_text,
            sem_dedupe_config_yaml_path=sem_dedupe_config_yaml_path,
        )
        print("********************* Generating Statistics *********************")
        print(f"After semantic dedupe for text files: {len(semantic_dataset_text.df)}")

        print("Executing the fuzzy dedupe pipeline...")
        CACHE_DIR = os.path.join(SCRIPT_DIR_PATH, "cache", "fuzzy_dedupe", "text")
        rm_dir(CACHE_DIR)
        fuzzy_dataset_text = fuzzy_dedupe(
            dataset=semantic_dataset_text, cache_dir=CACHE_DIR
        )
        CACHE_DIR = os.path.join(SCRIPT_DIR_PATH, "cache", "fuzzy_dedupe", "code")
        rm_dir(CACHE_DIR)
        fuzzy_dataset_code = fuzzy_dedupe(dataset=gpu_dataset_code, cache_dir=CACHE_DIR)

        dataset_text.df = fuzzy_dataset_text.df.to_backend("pandas")
        dataset_code.df = fuzzy_dataset_code.df.to_backend("pandas")
        print("********************* Generating Statistics *********************")
        print(f"After fuzzy dedupe for text files: {len(dataset_text.df)}")
        print(f"After fuzzy dedupe for code files: {len(dataset_code.df)}")

    final_dataset_text = dataset_text.persist()
    final_dataset_code = dataset_code.persist()

    print("Writing the results to disk...")

    # Overwrite existing files in the curated directory.
    out_path = os.path.join(DATA_DIR, "curated")

    if os.path.isdir(out_path):
        shutil.rmtree(out_path)

    os.makedirs(out_path)
    final_dataset_text.to_json(out_path, write_to_filename=True)
    final_dataset_code.to_json(out_path, write_to_filename=True)

    print("Writing results to disk completed")

    # Split the dataset by file category and save curated files (optional - to create blended datasets)
    print("Split dataset by metadata")
    separated_data_text = separate_by_metadata(
        final_dataset_text.df, out_path, "category"
    ).compute()
    separated_data_code = separate_by_metadata(
        final_dataset_code.df, out_path, "category"
    ).compute()

    client.close()


def blend_and_shuffle(
    args: Any, dataset_paths: list, dataset_weights: list, target_size: int
) -> None:
    """
    Blend and shuffle curated data based on file paths for continued pre-training

    Args:
        args (Any): Command-line arguments.
        dataset_paths (list): List containing directory paths where the different JSONL files are stored.
        dataset_weights (list): List setting weights for each directory path
        target_size (int): Target number of data samples after blending
    """
    root_path = os.path.join(DATA_DIR, "curated")
    output_path = root_path + "/data_blended"
    if os.path.isdir(output_path):
        shutil.rmtree(output_path)
    os.makedirs(output_path)

    # Blend the datasets
    datasets = [DocumentDataset.read_json(path) for path in dataset_paths]
    blended_dataset = nc.blend_datasets(target_size, datasets, dataset_weights)

    shuffle = nc.Shuffle(seed=42)
    blended_dataset = shuffle(blended_dataset)

    # Save the blend
    blended_dataset.to_json(output_path)


def main():
    parser = argparse.ArgumentParser()
    args = ArgumentHelper(parser).add_distributed_args().parse_args()
    # Limit the total number of workers to ensure we don't run out of memory.
    args.n_workers = min(args.n_workers, 8)
    print("Args: ", args)

    # Download all the sources and get the list of text and code files.
    text_files, code_files = download_sources(100, 100, 100)
    run_curation_pipeline(args, text_files, code_files)
    print("Data Curation completed")

    # blend and shuffle datasets
    root_path = os.path.join(DATA_DIR, "curated")
    dataset_paths = [
        root_path + "/CPP",
        root_path + "/VerilogVHDL",
        root_path + "/text",
        root_path + "/Python",
    ]
    dataset_weights = [1.0, 4.0, 4.0, 1.0]
    target_size = 20

    blend_and_shuffle(args, dataset_paths, dataset_weights, target_size)
    print("Data Blending completed")


if __name__ == "__main__":
    main()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import dask.dataframe as dd
import numpy as np
from dask import delayed

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modules.base import BaseModule
from nemo_curator.utils.module_utils import count_digits


class AddId(BaseModule):
    def __init__(
        self, id_field, id_prefix: str = "doc_id", start_index: Optional[int] = None
    ) -> None:
        super().__init__(input_backend="any")
        self.id_field = id_field
        self.id_prefix = id_prefix
        self.start_index = start_index

    def call(self, dataset: DocumentDataset) -> DocumentDataset:
        if self.start_index is None:
            return self._add_id_fast(dataset)
        else:
            return self._add_id_ordered(dataset)

    def _add_id_fast(self, dataset: DocumentDataset) -> DocumentDataset:
        meta = dataset.df._meta.copy()
        meta[self.id_field] = "string"
        meta[self.id_field] = meta[self.id_field].astype("string")

        partition_zero_padding = count_digits(dataset.df.npartitions)
        id_df = dataset.df.map_partitions(
            self._add_id_fast_partition,
            partition_zero_padding,
            meta=meta,
            enforce_metadata=False,
        )

        return DocumentDataset(id_df)

    def _add_id_fast_partition(self, partition, global_padding, partition_info=None):
        local_padding = count_digits(len(partition))
        global_id = partition_info["number"]

        id_column = [
            f"{self.id_prefix}-{local_id:0{local_padding}d}{global_id:0{global_padding}d}"
            for local_id in range(len(partition))
        ]
        partition[self.id_field] = id_column
        partition[self.id_field] = partition[self.id_field].astype("string")

        return partition

    def _add_id_ordered(self, dataset: DocumentDataset) -> DocumentDataset:
        original_meta = dataset.df._meta.copy()
        original_meta[self.id_field] = "string"
        original_meta[self.id_field] = original_meta[self.id_field].astype("string")
        delayed_dataset = dataset.df.to_delayed()

        parition_lengths = [0]
        for partition in delayed_dataset[:-1]:
            parition_lengths.append(delayed(len)(partition))

        lower_id_bounds = delayed(np.cumsum)(parition_lengths)
        delayed_id_dataset = []
        for i, partition in enumerate(delayed_dataset):
            delayed_id_dataset.append(
                delayed(self._add_id_ordered_partition)(partition, lower_id_bounds[i])
            )

        id_dataset = DocumentDataset(
            dataset_df=dd.from_delayed(delayed_id_dataset, meta=original_meta)
        )

        return id_dataset

    def _add_id_ordered_partition(self, partition, partition_start_id):
        id_column = [
            f"{self.id_prefix}-{int(i + self.start_index):010d}"
            for i in range(partition_start_id, len(partition) + partition_start_id)
        ]
        partition[self.id_field] = id_column
        partition[self.id_field] = partition[self.id_field].astype("string")

        return partition



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Literal

import pandas as pd
import pytest
from dask import dataframe as dd

from nemo_curator.utils.duplicates_removal import remove_duplicates


@pytest.fixture()
def ids():
    # Dataset has id a0...a9, b0...b9, c0...c9, d0...d9
    l = [f"{group}{i}" for group in ["a", "b", "c", "d"] for i in range(10)]
    return l


@pytest.fixture
def sample_data(ids):
    df = pd.DataFrame(
        {
            "id": ids,
            "text": [f"text for {_id}" for _id in ids],
        }
    )
    return dd.from_pandas(df, npartitions=4)


@pytest.fixture
def duplicate_data(ids):
    # In each group we want to keep only the first occurrence (e.g. a1, b1, c1, d1)
    df = pd.DataFrame([{"id": _id, "group": _id[0]} for _id in ids])
    return dd.from_pandas(df, npartitions=2)


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param("cudf", marks=pytest.mark.gpu),
    ],
)
@pytest.mark.parametrize("perform_shuffle", [False, True])
def test_remove_duplicates_basic(
    backend: Literal["cudf", "pandas"],
    perform_shuffle: bool,
    sample_data: dd.DataFrame,
    duplicate_data: dd.DataFrame,
):
    if perform_shuffle:
        # We shuffle the data to make sure that duplicates are not in the same partition
        duplicate_data = duplicate_data.sample(frac=1).reset_index(drop=True)

    sample_data = sample_data.to_backend(backend)
    duplicate_data = duplicate_data.to_backend(backend)

    # Test basic duplicate removal functionality
    result = remove_duplicates(
        left=sample_data,
        duplicates=duplicate_data,
        id_field="id",
        group_field="group",
        perform_shuffle=perform_shuffle,
    ).to_backend("pandas")

    result = result.compute()

    assert list(result.columns) == ["id", "text"]
    assert len(result) == 4
    # It's not guaranteed that we'll have a0, b0, c0, d0 in the result
    # So we should check the first character
    assert set(result["id"].apply(lambda x: x[0]).tolist()) == set(["a", "b", "c", "d"])


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param("cudf", marks=pytest.mark.gpu),
    ],
)
@pytest.mark.parametrize("perform_shuffle", [False, True])
def test_remove_duplicates_all_duplicates(
    backend: Literal["cudf", "pandas"],
    perform_shuffle: bool,
    ids: list[str],
    sample_data: dd.DataFrame,
):

    duplicates = dd.from_pandas(
        pd.DataFrame({"id": ids, "group": [1] * len(ids)}), npartitions=2
    )
    sample_data = sample_data.to_backend(backend)
    duplicates = duplicates.to_backend(backend)

    result = remove_duplicates(
        left=sample_data,
        duplicates=duplicates,
        id_field="id",
        group_field="group",
        perform_shuffle=perform_shuffle,
    ).to_backend("pandas")

    assert list(result.columns) == ["id", "text"]
    result = result.compute()
    if perform_shuffle:
        assert len(result) == 1
    else:
        # If we don't shuffle, and both partitions have the same group
        # in both partitions we'd be left with 1 row after "deduplication"
        # and after the left-anti join we'd be left with 2 rows
        assert len(result) == 2


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param("cudf", marks=pytest.mark.gpu),
    ],
)
@pytest.mark.parametrize("perform_shuffle", [False, True])
def test_not_remove_duplicates_unique(
    backend: Literal["cudf", "pandas"],
    perform_shuffle: bool,
    ids: list[str],
    sample_data: dd.DataFrame,
):
    # We create a dataset where first 30 ids are in one group
    # Next 9 ids are in distinct groups
    # And last id is not mentioned in duplicates

    duplicates = dd.from_pandas(
        pd.DataFrame(
            {
                "id": ids[:30] + ids[30:39],
                "group": ["group0"] * 30 + [f"group{i}" for i in range(1, 10)],
            }
        ),
        npartitions=2,
    )
    sample_data = sample_data.to_backend(backend)
    duplicates = duplicates.to_backend(backend)
    if perform_shuffle:
        # We shuffle the data to make sure that duplicates are not in the same partition
        duplicates = duplicates.sample(frac=1, random_state=42).reset_index(drop=True)

    result = remove_duplicates(
        left=sample_data,
        duplicates=duplicates,
        id_field="id",
        group_field="group",
        perform_shuffle=perform_shuffle,
    ).to_backend("pandas")

    result = result.compute()
    assert list(result.columns) == ["id", "text"]
    if perform_shuffle:
        # Since we've performed a shuffle, we know groups are collacated and there are 3 groups
        # 1.  1 row from the first group of 30
        # 2.  9 rows from the 9 distinct groups
        # 3. And 1 row from the last group which is not included in set of duplicates
        assert len(result) == 1 + 9 + 1
        # The last 10 ids should be in the result, there would be one more from the first 30
        assert set(ids[30:]).issubset(set(result["id"].tolist()))
    else:
        # If we don't shuffle, we'de be left with 2 partitions both having rows from group 1
        assert len(result) == 2 + 9 + 1


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param("cudf", marks=pytest.mark.gpu),
    ],
)
@pytest.mark.parametrize("left_npartitions", [2, 3])
@pytest.mark.parametrize("right_npartitions", [2, 3])
def test_remove_duplicates_repartition(
    backend: Literal["cudf", "pandas"],
    left_npartitions: int,
    right_npartitions: int,
):
    # Create sample dataframes with specific partition counts
    df1 = dd.from_pandas(
        pd.DataFrame({"id": ["a1", "a2", "a3"], "text": ["text1", "text2", "text3"]}),
        npartitions=left_npartitions,
    )  # dataset with 2 partitions

    duplicates = dd.from_pandas(
        pd.DataFrame(
            {"id": ["a1", "a2", "a3"], "group": ["group1", "group1", "group1"]}
        ),
        npartitions=right_npartitions,
    )  # duplicates dataset with 3 partitions
    df1 = df1.to_backend(backend)
    duplicates = duplicates.to_backend(backend)

    # Test that it raises ValueError when right npartitions are greater than left npartitions
    output = remove_duplicates(
        left=df1,
        duplicates=duplicates,
        id_field="id",
        group_field="group",
    )
    output_dask_graph_keys = set(
        k[0].rsplit("-", 1)[0] for k in output.optimize().__dask_graph__().keys()
    )
    assert "broadcastjoin" in output_dask_graph_keys
    if left_npartitions < right_npartitions:
        assert "repartitiontofewer" in output_dask_graph_keys
    else:
        assert "repartitiontofewer" not in output_dask_graph_keys



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import time

from retriever_hardnegative_miner import HardNegativeMiner

from config.config import RetrieverHardNegativeMiningConfig
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.file_utils import get_all_files_paths_under


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input-dir",
        type=str,
        default="",
        help="Input dir path containing annotated data files in jsonl format",
    )
    parser.add_argument(
        "--hard-negative-mining-config",
        type=str,
        default="",
        help="Configuration yaml file path containing config for hard negative mining",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        default="",
        help="Output file containing clustered dataset",
    )
    parser.add_argument(
        "--api-key",
        type=str,
        default=None,
        help="The API key to use for the synthetic data generation LLM client.",
    )
    parser.add_argument(
        "--api-timeout",
        type=int,
        default=120,
        help="The timeout value for API calls in seconds.",
    )
    args = parser.parse_args()

    if not os.path.exists(args.input_dir):
        raise ValueError("Input directory not found")

    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    elif not any(os.scandir(args.output_dir)):
        print("Provided directory exists but is empty, using the empty directory")
    else:
        raise ValueError("Output directory exists already, use a new directory!")

    if args.input_dir:
        input_files = get_all_files_paths_under(args.input_dir, keep_extensions="jsonl")
        input_dataset = DocumentDataset.read_json(input_files)
    else:
        raise ValueError("provide input file path")
    if args.hard_negative_mining_config:
        cfg = RetrieverHardNegativeMiningConfig.from_yaml(
            args.hard_negative_mining_config
        )
    else:
        raise ValueError("provide config for hard negative mining")
    if args.api_key:
        cfg.api_key = args.api_key
    if cfg.cluster_output_dir:
        if not os.path.exists(cfg.cluster_output_dir):
            os.makedirs(cfg.cluster_output_dir)
    else:
        cfg.cluster_output_dir = os.path.join(args.output_dir, "clusters")
    if cfg.logger_output_dir:
        if not os.path.exists(cfg.logger_output_dir):
            os.makedirs(cfg.logger_output_dir)
    else:
        cfg.logger_output_dir = os.path.join(args.output_dir, "logs")

    st_time = time.time()
    miner = HardNegativeMiner(cfg)
    clustered_dataset = miner.repartition_semantic_similarity(input_dataset)
    clustered_dataset.persist()

    # saving clustered dataset
    print("saving clustered dataset")
    clustered_dataset.df.to_json(args.output_dir)
    print("Time taken to cluster data = {:.2f} s".format(time.time() - st_time))


if __name__ == "__main__":
    dask_client = get_client(cluster_type="gpu")
    main()
